__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import sys
import pandas as pd
sys.setrecursionlimit(15000000)
from umiche.util.Console import Console


class Directional:

    def __init__(
            self,
            heterogeneity=False,
            verbose=True,
    ):
        self.heterogeneity = heterogeneity

        self.console = Console()
        self.console.verbose = verbose

    def umi_tools(
            self,
            connected_components,
            df_umi_uniq_val_cnt,
            graph_adj,
    ):
        """

        Parameters
        ----------
        connected_components
            {0: ['A', 'B', 'C', 'D', 'E', 'F']}
        df_umi_uniq_val_cnt
            A    456
            E     90
            D     72
            B      2
            C      2
            F      1
            dtype: int64
        graph_adj
            {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}
        Returns
        -------

        """
        cc_sub_cnt = []
        cc_subs = {}
        cc_apvs = {}
        cc_disapvs = {}
        for i, cc in connected_components.items():

            cc_sub, apv_node_nbr, disapv_node_nbr = self.umi_tools_(
                df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
                cc=cc,
                graph_adj=graph_adj,
            )
            from umiche.deduplicate.method.ReformKit import ReformKit as refkit
            # print([*cc_sub.values()])
            # if len([*cc_sub.values()]) > 1:
            #     print([*cc_sub.values()])
                # print(graph_adj)

            if cc == [23, 41, 27, 62, 85, 48, 58, 81, 28, 46, 71, 45, 38, 39, 64, 49]:
            # if cc == [69, 72, 838, 1221, 97, 210, 249, 315, 324, 374, 457, 658, 727, 760, 771, 933, 1073, 1126, 1198, 1260, 1271, 1307, 1498, 1505, 1541, 1563, 914, 946, 1083, 684, 1288, 1543, 822, 1174, 119, 290, 303, 1315, 204, 218, 289, 302, 404, 545, 586, 633, 674, 709, 720, 802, 884, 943, 980, 1355, 1436, 1488, 1553, 786, 1549, 537, 867, 1649, 1255, 701, 1080, 347, 251]:
                print(cc)
                print(cc_sub)
                print([*cc_sub.values()])

                from umiche.path import to
                df_umi_uniq_val_cnt.to_csv(to('data/simu/mclumi/seq_errs/umi_uniq_val_cnt.txt'), sep='\t', index=True, header=None)
                print('break')
                break
            # print(cc_sub)
            cc_sub_cnt.append(len(cc_sub))
            cc_subs['cc_' + str(i)] = cc_sub
            cc_apvs['cc_' + str(i)] = apv_node_nbr
            cc_disapvs['cc_' + str(i)] = disapv_node_nbr
        # print(sum(cc_sub_cnt))
        # print(cc_subs)
        # print(cc_apvs)
        # print(cc_disapvs)
        if self.heterogeneity:
            return (
                sum(cc_sub_cnt),
                cc_subs,
                cc_apvs,
                cc_disapvs,
            )
        else:
            return {
                "count": sum(cc_sub_cnt),
                "clusters": cc_subs,
                "apv": cc_apvs,
                "disapv": cc_disapvs,
            }

    def umi_tools_(
            self,
            df_umi_uniq_val_cnt,
            cc,
            graph_adj,
    ):
        """

        Parameters
        ----------
        df_umi_uniq_val_cnt
            A    456
            E     90
            D     72
            B      2
            C      2
            F      1
            dtype: int64
        cc
            {0: ['A', 'B', 'C', 'D', 'E', 'F']}
        graph_adj
            {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}

        Returns
        -------

        """
        cc_node_sorted = df_umi_uniq_val_cnt.loc[df_umi_uniq_val_cnt.index.isin(cc)].sort_values(ascending=False).to_dict()
        ### @@ cc_umi_sorted
        # {'A': 456, 'E': 90, 'D': 72, 'B': 2, 'C': 2, 'F': 1}
        nodes = [*cc_node_sorted.keys()]
        # print(nodes)
        ### @@ cc_sorted
        # ['A', 'E', 'D', 'B', 'C', 'F']
        node_cp = nodes.copy()
        node_set_remaining = set(node_cp)
        ### @@ node_set_remaining
        # {'C', 'F', 'E', 'B', 'D', 'A'}
        cc_sub = {}
        apv_node_nbr = {}
        disapv_node_nbr = {}
        while nodes:
            e = nodes.pop(0)
            if e in node_set_remaining:
                seen, apv, disapv = self.dfs(
                    node=e,
                    node_val_sorted=cc_node_sorted,
                    node_set_remaining=node_set_remaining,
                    graph_adj=graph_adj,
                )
                ### @@ e, seen
                # A {'C', 'D', 'F', 'A', 'B'}
                # E {'E'}
                cc_sub['node_' + str(e)] = list(seen)
                apv_node_nbr['node_' + str(e)] = apv
                disapv_node_nbr['node_' + str(e)] = disapv
                node_set_remaining = node_set_remaining - seen
                self.console.print('remaining: {}'.format(node_set_remaining))
                self.console.print('disapproval {}'.format(disapv))
                ### @@ print('disapproval {}'.format(disapv))
                # disapproval []
                # disapproval [[183, 103]]
                # disapproval [[131, 4], [131, 147]]
                # ...
                # disapproval [[133, 194]]
                # disapproval []
            else:
                continue
        ### @@ disapv_node_nbr
        # {'node_0': []}
        # {'node_36': [[183, 103]]}
        # {'node_29': [[131, 4], [131, 147]], 'node_4': []}
        # {'node_7': []}
        # {'node_28': [[8, 57]]}
        # ...
        # {'node_59': [[133, 194]]}
        # {'node_63': []}
        return cc_sub, apv_node_nbr, disapv_node_nbr

    def dfs(
            self,
            node,
            node_val_sorted,
            node_set_remaining,
            graph_adj,
    ):
        """

        Parameters
        ----------
        node
            'A'
        node_val_sorted
            A    456
            E     90
            D     72
            B      2
            C      2
            F      1
            dtype: int64
        node_set_remaining
            {'F', 'A', 'D', 'C', 'E', 'B'}
            {'E'}
        graph_adj
            {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}
        Returns
        -------

        """
        visited = set()
        approval = []
        disapproval = []
        g = graph_adj
        def search(node):
            visited.add(node)
            self.console.print('======>visited UMI nodes: {}'.format(visited))
            self.console.print(visited)
            for neighbor in g[node]:
                self.console.print('=========>the neighbor: {}'.format(neighbor))
                if neighbor not in visited:
                    if neighbor in node_set_remaining:
                        if node_val_sorted[node] >= 2 * node_val_sorted[neighbor] - 1:
                            approval.append([node, neighbor])
                            search(neighbor)
                        else:
                            disapproval.append([node, neighbor])
        search(node)
        ### @@ approval
        # {'cc_0': {'node_A': [['A', 'B'], ['A', 'C'], ['A', 'D'], ['D', 'F']], 'node_E': []}}
        ### @@ disapproval
        # {'cc_0': {'node_A': [['B', 'C'], ['D', 'E']], 'node_E': []}}
        return visited, approval, disapproval

    def decompose(
            self,
            cc_sub_dict,
    ):
        """

        Parameters
        ----------
        cc_sub_dict
            input: {'cc_0': {'node_A': ['C', 'A', 'D', 'B', 'F'], 'node_E': ['E']}}
        Returns
        -------
            output: {0: ['C', 'A', 'D', 'B', 'F'], 1: ['E']}

        """
        cc_cnt = 0
        ccs = {}
        for k1, v1 in cc_sub_dict.items():
            for k2, v2 in v1.items():
                ccs[cc_cnt] = v2
                cc_cnt += 1
        return ccs


if __name__ == "__main__":
    import pandas as pd
    from umiche.deduplicate.method.Cluster import Cluster as umiclust

    p = Directional()

    ### @@ data from UMI-tools
    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D'],
    #     'F': ['D'],
    # }
    # print("An adjacency list of a graph:\n{}".format(graph_adj))
    #
    # node_val_sorted = pd.Series({
    #     'A': 456,
    #     'E': 90,
    #     'D': 72,
    #     'B': 2,
    #     'C': 2,
    #     'F': 1,
    # })
    # print("Counts sorted:\n{}".format(node_val_sorted))

    ### @@ data from mine
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }
    print("An adjacency list of a graph:\n{}".format(graph_adj))

    node_val_sorted = pd.Series({
        'A': 120,
        'D': 90,
        'E': 50,
        'G': 5,
        'B': 2,
        'C': 2,
        'F': 1,
    })
    print("Counts sorted:\n{}".format(node_val_sorted))

    ccs = umiclust().cc(graph_adj=graph_adj)
    print("Connected components:\n{}".format(ccs))

    dedup_res = p.umi_tools(
        connected_components=ccs,
        df_umi_uniq_val_cnt=node_val_sorted,
        graph_adj=graph_adj
    )
    dedup_count = dedup_res['count']
    dedup_clusters = dedup_res['clusters']
    print("deduplicated count:\n{}".format(dedup_count))
    print("deduplicated clusters:\n{}".format(dedup_clusters))

    dedup_clusters_dc = p.decompose(dedup_clusters)
    print("deduplicated clusters decomposed:\n{}".format(dedup_clusters_dc))

    print(dedup_res['apv'])
    print(dedup_res['disapv'])
    # {'cc_0': {'node_A': [['A', 'B'], ['A', 'C']], 'node_D': [['D', 'F']], 'node_E': [['E', 'G']]}}
    # {'cc_0': {'node_A': [['B', 'C'], ['A', 'D']], 'node_D': [['D', 'E'], ['F', 'G']], 'node_E': []}}



import numpy as np


class variance(object):

    def __init__(self, gc_mat):
        self.gc_mat = gc_mat

    def pergene(self, ):
        return self.gc_mat.var(axis=0)

    def percell(self, ):
        return self.gc_mat.var(axis=1)

    def all(self, ):
        return self.gc_mat.var()


if __name__ == "__main__":
    gc_mat = np.array([
        [3., 3., 3., 6., 6.],
        [1., 1., 1., 2., 2.],
        [1., 22., 1., 2., 2.]
    ])
    p = variance(gc_mat)
    print(p.pergene())
    print(p.percell())
    print(p.all())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import textwrap
import numpy as np
import pandas as pd
from simreadflow.util.sequence.fastq.Read import read as rfastq
from simreadflow.util.random.Number import number as rannum
from umiche.trim.Template import template as umitrim
from umiche.trim.Reader import reader as trimreader
from simreadflow.util.file.read.Reader import reader as gfreader
from simreadflow.simulate.dispatcher.batch.UMI import umi as generalstarter
from simreadflow.read.similarity.distance.Hamming import hamming
from simreadflow.util.file.write.Writer import writer as fwriter
from umiche.path import to


class Shakeup(generalstarter):

    def __init__(self, umi_ref_fpn, fastq_fp, cat):
        super(Shakeup, self).__init__()
        self.umi_ref_fpn = umi_ref_fpn
        self.fastq_fp = fastq_fp
        self.cat = cat
        self.umitrim = umitrim
        self.gfreader = gfreader()
        self.rfastq = rfastq
        self.trimreader = trimreader()
        self.rannum = rannum()
        self.fwriter = fwriter()
        self.umi_raw = self.gfreader.generic(
            df_fpn=self.umi_ref_fpn
        )[0].values
        self.umi_raw = pd.DataFrame.from_dict({i: e for i, e in enumerate(self.umi_raw)}, orient='index')
        self.umi_mono_raw = self.umi_raw[0].apply(lambda x: ''.join([i[0] for i in textwrap.wrap(x, 2)])).to_dict()
        print(self.umi_mono_raw)

    def rea(self, ):
        df_stat = pd.DataFrame()
        for id, i_seq_err in enumerate(self.seq_errs):
            read_stime = time.time()
            names, seqs, _, _ = self.rfastq().fromgz(
                fastq_path=self.fastq_fp,
                fastq_name=self.cat + '_' + str(id),
                method='pyfastx',
            )
            print('===>file read time: {:.3f}s'.format(time.time() - read_stime))
            df_fastq = self.trimreader.todf(names=names, seqs=seqs)
            mono_corr_stime = time.time()
            df_fastq['umi_mono_corr'] = df_fastq['umi'].apply(lambda x: self.correct(x))
            print('===>mono_corr time: {:.3f}s'.format(time.time() - mono_corr_stime))
            hm_stime = time.time()
            df_stat['umi_hm' + str(id)] = df_fastq.apply(lambda x: hamming().general(
                s1=x['umi_mono_corr'],
                s2=self.umi_mono_raw[x['umi#']],
            ), axis=1)
            print('===>hamming time: {:.3f}s'.format(time.time() - hm_stime))
            tui = df_stat['umi_hm' + str(id)].values
            print(len(tui[tui != 0]))
        # self.fwriter.generic(df=df_stat, sv_fpn=to('data/simu/umi/seq_errs/dimer/trimmed/dasd.txt'), df_sep='\t')
        return

    def correct(self, umi):
        umi_dimers = textwrap.wrap(umi, 2)
        t = []
        # print(umi_dimers)
        for umi_dimer in umi_dimers:
            if umi_dimer[0] != umi_dimer[1]:
                rand_index = np.random.randint(low=0, high=2, size=1)[0]
                # print(rand_index)
                t.append(umi_dimer[rand_index])
            else:
                t.append(umi_dimer[0])
        return ''.join(t)


if __name__ == "__main__":
    p = Shakeup(
        umi_ref_fpn=to('data/simu/umi/seq_errs/dimer/umi.txt'),
        fastq_fp=to('data/simu/umi/seq_errs/dimer/trimmed/'),
        cat='seq_err',
    )
    print(p.rea())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import yaml
from umiche.util.Console import Console


class Parameter:

    def __init__(
            self,
            param_fpn,
            verbose=True,
    ):
        self.console = Console()
        self.console.verbose = verbose
        with open(param_fpn, "r") as f:
            self.params = yaml.safe_load(f)
            for i, (k, item) in enumerate(self.params.items()):
                self.console.print("======>key {}: {}".format(i+1, k))
                self.console.print("=========>value: {}".format(item))

    @property
    def fixed(self, ):
        return self.params['fixed']

    @property
    def varied(self, ):
        return self.params['varied']

    @property
    def trimmed(self, ):
        return self.params['trimmed']

    @property
    def dedup(self, ):
        return self.params['dedup']

    @property
    def work_dir(self, ):
        return self.params['work_dir']

    @property
    def anchor(self, ):
        return self.params['anchor']

    @property
    def fastq_fp(self, ):
        return self.params['fastq_fp']

    @property
    def umi_ref_fpn(self, ):
        return self.params['umi_ref_fpn']

    @property
    def file_names(self, ):
        return {
            'pcr_nums': 'pcr_num_',
            'pcr_errs': 'pcr_err_',
            'seq_errs': 'seq_err_',
            'ampl_rates': 'ampl_rate_',
            'umi_lens': 'umi_len_',
            'seq_deps': 'seq_dep_',
            'umi_nums': 'umi_num_',
        }


if __name__ == "__main__":
    p = Parameter(
        param_fpn='./params/param_fpn.txt'
    )
    print(p.file_names)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Union
import numpy as np

from umiche.deduplicate.TMR import TMR
from umiche.deduplicate.method.trimer.Collapse import Collapse



def collapse():
    return Collapse()


def tmr(
        error_rate: Union[float, np.ndarray] = 0.00001,
):
    return TMR(
        error_rate=error_rate,
    )


if __name__ == "__main__":
    splitter = collapse()
    # print(splitter.split_by_mv('AAACCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))

    # print(splitter.split_by_mv('AAATCCGGATTTCGGAAATTTGGGCACCCC', recur_len=3))
    # print(splitter.split_by_mv('AAATCCGGATTTCGGAAATTTGGGCCCCCC', recur_len=3))
    # print(splitter.split_by_mv('AAATCCGGATTTGGGAAATTTGGGCCCCCC', recur_len=3))
    # print(splitter.split_by_mv('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))

    # print(splitter.split_to_all('AAATCCGGATTTCGGAAATTTGGGCACCCC', recur_len=3))
    # print(splitter.split_to_all('AAATCCGGATTTCGGAAATTTGGGCCCCCC', recur_len=3))
    # print(splitter.split_to_all('AAATCCGGATTTGGGAAATTTGGGCCCCCC', recur_len=3))
    # print(splitter.split_to_all('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))

    # print(splitter.majority_vote('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))
    # print(splitter.take_by_order('AAATCCGGGTTTGGGAAATTTGGGCCCCCC', recur_len=3))

    # print(splitter.vote('AAA', recur_len=3))
    # print(splitter.vote('TAA', recur_len=3))
    # print(splitter.vote('TGA', recur_len=3))

    p = tmr(error_rate=0.00001)
    # print(p.get_num_nt())
    # print(p.homotrimer_block_error)
    # print(p.bitflip_block_error)
    # print(p.homotrimer_umi_error)
    # p.set_num_nt(2)
    # # print(p.get_num_nt())
    # print(p.homotrimer_umi_error)
    # print(p.homotrimer_umi_error_free)
    p.set_num_nt(12)
    print(p.homotrimer_umi_error)
    print(p.bitflip_code_error)
    print(p.monomer_umi_error)
    # print(p.homotrimer_umi_error_free)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import pandas as pd

from umiche.bam.Reader import Reader as alireader

from umiche.bam.Build import Build as umibuild
from umiche.deduplicate.Gadgetry import Gadgetry as umigadgetry
from umiche.deduplicate.Tabulate import Tabulate as umitab

# dedup methods
from umiche.deduplicate.method.Cluster import Cluster as umiclust

from umiche.util.Console import Console


class OnePos:

    def __init__(
            self,
            bam_fpn,
            ed_thres,
            umi_='_',
            work_dir='./',

            heterogeneity=False,
            verbose=False,

            **kwargs,
    ):
        """

        Parameters
        ----------
        bam_fpn
            str - the full path of a BAM file curated by requirements of different dedup modules
        ed_thres
            int - an edit distance threshold (>1, integer)
        method
            str - a deduplication method (mcl, mcl_val, mcl_ed, cluster, unique, ajacency, directional)
        mode
            str - externally or internally run the module (external by defualt, internal)
        mcl_fold_thres
            float - a mcl fold threshold (1.5 by defualt)
        inflat_val
            float - an inflation value for generating mcl clusters (2.0 by defualt)
        exp_val
            int - an expansion value for generating mcl clusters (2 by defualt)
        iter_num
            int - number of iterations for mcl (100 by defualt)
        sv_fpn
            str - the deduplication file path
        verbose
            bool - print log on the console, (True by default or False)
        """
        self.bam_fpn = bam_fpn
        self.ed_thres = ed_thres
        self.work_dir = work_dir
        self.heterogeneity = heterogeneity
        self.verbose = verbose
        self.kwargs = kwargs
        # print("params in OnePos: {}".format(self.kwargs))

        self.umiclust = umiclust()

        self.umibuild = umibuild
        self.umigadgetry = umigadgetry()
        self.console = Console()
        self.console.verbose = self.verbose

        # sys.stdout = open(self.work_dir + self.method + '_log.txt', 'w')

        self.alireader = alireader(bam_fpn=self.bam_fpn, verbose=self.verbose)
        self.df_bam = self.alireader.todf(tags=[])
        ### @@ self.df_bam
        #           id  ... query_qualities
        # 0          0  ...            None
        # 1          1  ...            None
        # 2          2  ...            None
        # ...      ...  ...             ...
        # 20683  20683  ...            None
        self.console.print('======># of raw reads: {}'.format(self.df_bam.shape[0]))
        self.df_bam = self.df_bam.loc[self.df_bam['reference_id'] != -1]
        self.console.print('======># of reads with qualified chrs: {}'.format(self.df_bam.shape[0]))

        self.df_bam['umi'] = self.df_bam['query_name'].apply(lambda x: x.split(umi_)[-1])
        # print( self.df_bam['umi'])
        self.console.print('======># of unique umis: {}'.format(self.df_bam['umi'].unique().shape[0]))
        self.console.print('======># of redundant umis: {}'.format(self.df_bam['umi'].shape[0]))
        self.console.print('======>edit distance thres: {}'.format(self.ed_thres))

        self.df_bam['source'] = 1
        ### @@ self.df_bam
        #           id                     query_name  ...        umi  source
        # 0          0   SRR2057595.2985267_ACCGGTTTA  ...  ACCGGTTTA       1
        # 1          1  SRR2057595.13520751_CCAGGTTCT  ...  CCAGGTTCT       1
        # 2          2   SRR2057595.8901432_AGCGGTTAC  ...  AGCGGTTAC       1
        # ...      ...                            ...  ...        ...     ...
        # 20683  20683  SRR2057595.11966225_ACCGGTTGG  ...  ACCGGTTGG       1
        # [20684 rows x 13 columns]
        self.df_bam_gp = self.df_bam.groupby(by=['source'])
        self.pos_gp_keys = self.df_bam_gp.groups.keys()

        self.console.print('======># of columns in the bam df: {}'.format(len(self.df_bam.columns)))
        self.console.print('======>Columns in the bam df: {}'.format(self.df_bam.columns.tolist()))
        self.console.print('======># of raw reads:\n{}'.format(self.df_bam))

        self.console.print('===>start building umi graphs...')
        umi_graph_build_stime = time.time()
        pos_gps = []
        res_sum = []
        for pos_g in self.pos_gp_keys:
            umi_vignette = self.umibuild(
                df=self.df_bam_gp.get_group(pos_g),
                ed_thres=self.ed_thres,
                verbose=verbose,
            ).data_summary
            # print(umi_vignette)
            # if len(umi_vignette['int_to_umi_dict']) == 1:
            #     print(True)
            #     print(umi_vignette['int_to_umi_dict'])
            #     continue
            # else:
            cc = self.umiclust.cc(umi_vignette['graph_adj'])
            # import json
            # with open('data.json', 'w') as f:
            #     json.dump(cc, f)
            pos_gps.append(pos_g)
            res_sum.append([
                umi_vignette,
                cc,
                umi_vignette['ave_ed'],
                [*umi_vignette['int_to_umi_dict'].keys()],
            ])
            ### @@ self.df['uniq_repr_nodes'] or [*umi_vignette['int_to_umi_dict'].keys()]
            # [0, 1, 2, ..., 1948]
        self.df = pd.DataFrame(
            data=res_sum,
            columns=['vignette', 'cc', 'ave_ed', 'uniq_repr_nodes'],
            index=pos_gps,
        )
        ### @@  self.df
        # vignette  ...   uniq_repr_nodes
        # 1  {'graph_adj': {0: [77, 81, 97, 153, 205, 228, ...  ...  [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,...
        # [1 rows x 3 columns]
        self.console.print('===>time for building umi graphs: {:.2f}s'.format(time.time() - umi_graph_build_stime))

        self.df['num_uniq_umis'] = self.df['uniq_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        ### @@ self.df['num_uniq_umis']
        # 1    1949
        # Name: num_uniq_umis, dtype: int64
        # self.console.print('===>start deduplication by the {} method...'.format(self.method))
        # sys.stdout.close()

    def unique(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).unique()

    def cluster(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).cluster()

    def adjacency(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).adjacency()

    def directional(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).directional()

    def mcl(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl(
            **self.kwargs
        )

    def mcl_cc_all_node_umis(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_cc_all_node_umis(
            **self.kwargs
        )

    def mcl_val(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_val(
            **self.kwargs
        )

    def mcl_ed(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).mcl_ed(
            **self.kwargs
        )

    def dbscan_seq_onehot(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).clustering_umi_seq_onehot(
            clustering_method='dbscan',
            **self.kwargs
        )

    def birch_seq_onehot(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).clustering_umi_seq_onehot(
            clustering_method='birch',
            **self.kwargs
        )

    def hdbscan_seq_onehot(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).clustering_umi_seq_onehot(
            clustering_method='hdbscan',
            **self.kwargs
        )

    def aprop_seq_onehot(self, ) -> pd.DataFrame:
        return umitab(
            df=self.df,
            df_bam=self.df_bam,
            bam_fpn=self.bam_fpn,
            work_dir=self.work_dir,
            heterogeneity=self.heterogeneity,
            verbose=False,
        ).clustering_umi_seq_onehot(
            clustering_method='aprop',
            **self.kwargs
        )


if __name__ == "__main__":
    from umiche.path import to

    umiche = OnePos(
        # bam_fpn=to('data/example.bam'),
        # bam_fpn=to('data/example_bundle.bam'),
        bam_fpn=to('data/simu/mclumi/seq_errs/permute_0/trimmed/bam/seq_err_14.bam'),
        mcl_fold_thres=1.6,
        inflat_val=1.7,
        exp_val=2,
        iter_num=100,
        ed_thres=1,
        work_dir=to('data/'),

        verbose=False,  # False True
        heterogeneity=False,  # False True
    )

    # print(umiche.unique())
    # print(umiche.cluster())
    # print(umiche.adjacency())
    print(umiche.directional())
    # print(umiche.mcl())
    # print(umiche.mcl_val())
    # print(umiche.mcl_ed())



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import matplotlib.colors as mcolors


class Element:

    def __init__(self, ):
        self.color_list = self.color()

    def color(
            self,
            which='ccs4',
            is_random=False,
    ):
        """
        References
        ----------
            https://matplotlib.org/stable/gallery/color/named_colors.html

        It includes the color of the element
            'firebrick',
            'royalblue',
            'springgreen',
            'burlywood',
            'oldlace',
            'lightgoldenrodyellow',
            'grey',
            'cyan',
            'crimson',
            'mediumvioletred',
            'maroon',
            'mediumturquoise',
            'teal',
            'azure',
            'palevioletred',
            'mediumslateblue',
            'olivedrab',
            'darkred',
            'dimgrey',
            'lightgreen',
            'blueviolet',
            'chartreuse',
            'whitesmoke',
            'mediumspringgreen',
            'yellowgreen',
            'green',
            'darkorchid',
            'lawngreen',
            'black',
            'darkgoldenrod',
            'moccasin',
            'seashell',
            'orange',
            'slateblue',
            'hotpink',
            'tan',
            'darkolivegreen',
            'mediumorchid',
            'darkgreen',
            'navajowhite',
            'khaki',
            'paleturquoise',
            'darkgray',
            'darkorange',
            'salmon',
            'floralwhite',
            'lightgray',
            'darkblue',
            'lightslategray',
            'darkslateblue',
            'goldenrod',
            'lightcyan',
            'papayawhip',
            'wheat',
            'lightslategrey',
            'darksalmon',
            'skyblue',
            'violet',
            'coral',
            'mediumseagreen',
            'darkmagenta',
            'palegreen',
            'magenta',
            'lavenderblush',
            'darkkhaki',
            'darkturquoise',
            'lightsalmon',
            'blanchedalmond',
            'deepskyblue',
            'mediumaquamarine',
            'mintcream',
            'lightyellow',
            'linen',
            'peachpuff',
            'aliceblue',
            'gold',
            'ghostwhite',
            'silver',
            'bisque',
            'aquamarine',
            'lightblue',
            'lavender',
            'antiquewhite',
            'dimgray',
            'lightskyblue',
            'pink',
            'beige',
            'indianred',
            'rosybrown',
            'rebeccapurple',
            'lightgrey',
            'powderblue',
            'chocolate',
            'darkgrey',
            'purple',
            'darkslategrey',
            'cornsilk',
            'turquoise',
            'blue',
            'greenyellow',
            'red',
            'lime',
            'ivory',
            'royalblue',
            'palegoldenrod',
            'lightsteelblue',
            'slategray',
            'gray',
            'brown',
            'mistyrose',
            'cornflowerblue',
            'tomato',
            'indigo',
            'snow',
            'darkslategray',
            'lightpink',
            'darkviolet',
            'limegreen',
            'honeydew',
            'yellow',
            'sienna',
            'mediumblue',
            'fuchsia',
            'forestgreen',
            'orchid',
            'plum',
            'lemonchiffon',
            'gainsboro',
            'sandybrown',
            'slategrey',
            'mediumpurple',
            'saddlebrown',
            'lightcoral',
            'midnightblue',
            'navy',
            'thistle',
            'dodgerblue',
            'lightseagreen',
            'darkseagreen',
            'cadetblue',
            'seagreen',
            'darkcyan',
            'steelblue',
            'white',
            'peru',
            'deeppink',
            'olive',
            'firebrick',
            'aqua',
            'orangered',
        Returns
        -------

        """
        if which == 'base':
            colors = mcolors.BASE_COLORS
        elif which == 'ccs4':
            colors = mcolors.CSS4_COLORS
        elif which == 'tableau':
            colors = mcolors.TABLEAU_COLORS
        elif which == 'xkcd':
            colors = mcolors.XKCD_COLORS
        else:
            colors = mcolors.CSS4_COLORS
        # return np.random.permutation([*colors.keys()]).tolist()
        return [*colors.keys()]


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from umiche.util.Console import Console


class Writer:

    def __init__(
            self,
             df,
            verbose=False,
    ):
        import pysam

        self.df = df
        self.console = Console()
        self.console.verbose = verbose

    def tobam(
            self,
            tobam_fpn,
            tmpl_bam_fpn,
            whitelist=[],
    ):
        tmpl_bam = pysam.AlignmentFile(tmpl_bam_fpn, "rb")
        write_to_bam = pysam.AlignmentFile(tobam_fpn, "wb", template=tmpl_bam)
        fs = self.df.loc[self.df['id'].isin(whitelist)]['read']
        for i in fs:
            # print(i)
            write_to_bam.write(i)
        write_to_bam.close()
        return write_to_bam

    def tobam_trust4(
            self,
            tobam_fpn,
            # tmpl_bam_fpn,
            # whitelist=[],
    ):
        # 'wb' indicates that it's a binary file opened for writing
        # print(pysam.AlignmentHeader())
        # with pysam.AlignmentFile(tobam_fpn, 'wb', header=pysam.AlignmentHeader()) as output_bam:
        #     # Create a new alignment object for your query sequence
        #     # Replace placeholders with actual values
        #     # fs = self.df.loc[self.df['id'].isin(whitelist)]['read']
        #     # for i in fs:
        #         new_alignment = pysam.AlignedSegment()
        #         new_alignment.query_name = "YourQueryName"
        #         new_alignment.query_sequence = "ATCG"  # Replace with the actual sequence
        #         new_alignment.flag = 0x0  # Replace with appropriate flag
        #         new_alignment.reference_id = -1  # No reference ID specified
        #         new_alignment.reference_start = 0  # Replace with the start position
        #         new_alignment.mapping_quality = 60  # Replace with the mapping quality
        #         new_alignment.cigar = [(0, 4)]  # Replace with the CIGAR string
        #         new_alignment.next_reference_id = -1  # No next reference ID specified
        #         new_alignment.next_reference_start = 0  # Replace with the next reference start position
        #         new_alignment.template_length = 0  # Replace with the template length
        #         new_alignment.query_qualities = [30, 30, 30, 30]  # Replace with the quality scores
        #
        #         # Add the new alignment to the BAM file
        #         output_bam.write(new_alignment)
        header = {'HD': {'VN': '1.0'},
                  'SQ': [{'LN': 1575, 'SN': 'chr1'},
                         {'LN': 1584, 'SN': 'chr2'}]}

        with pysam.AlignmentFile(tobam_fpn, "wb", header=header) as outf:
            a = pysam.AlignedSegment()
            a.query_name = "read_28833_29006_6945"
            a.query_sequence = "AGCTTAGCTAGCTACCTATATCTTGGTCTTGGCCG"
            a.flag = 99
            a.reference_id = 0
            a.reference_start = 32
            a.mapping_quality = 20
            a.cigar = ((0, 10), (2, 1), (0, 25))
            a.next_reference_id = 0
            a.next_reference_start = 199
            a.template_length = 167
            a.query_qualities = pysam.qualitystring_to_array("<<<<<<<<<<<<<<<<<<<<<:<9/,&,22;;<<<")
            a.tags = (("NM", 1),
                      ("RG", "L1"))
            outf.write(a)
        return 1


if __name__ == "__main__":
    p = Writer(df=pd.DataFrame())

    p.tobam_trust4(
        tobam_fpn='./trimmed2.bam',
        # tmpl_bam_fpn=to('data/bone_marrow/merge_corrected_sorted_little.bam'),
        # whitelist=df.index,
    )



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

from typing import Union, List, Dict
import numpy as np

from umiche.plot.graph.Cluster import Cluster
from umiche.plot.scenario.DedupMultiple import DedupMultiple
from umiche.plot.scenario.DedupSingle import DedupSingle
from umiche.plot.scenario.DedupMultipleTrimer import DedupMultipleTrimer
from umiche.plot.line.TripletErrorCode import TripletErrorCode
from umiche.plot.line.Anchor import Anchor
from umiche.plot.scenario.DedupMultipleTrimerSetCover import DedupMultipleTrimerSetCover
from umiche.plot.scenario.TraceSingle import TraceSingle
from umiche.plot.scenario.TraceMultiple import TraceMultiple
from umiche.plot.scenario.Line import Line


def graph_cluster(
        dedup_cluster,
        ccs,
        edge_list,
        umi_uniq_val_cnt_fpn,
):
    return Cluster(
        dedup_cluster=dedup_cluster,
        ccs=ccs,
        edge_list=edge_list,
        umi_uniq_val_cnt_fpn=umi_uniq_val_cnt_fpn,
    )


def dedup_multiple(
        scenarios,
        methods,
        param_fpn,
):
    return DedupMultiple(
        scenarios=scenarios,
        methods=methods,
        param_fpn=param_fpn,
    )


def dedup_single(
        df_dedup,
        df_dedup_perm_melt,
):
    return DedupSingle(
        df=df_dedup,
        df_melt=df_dedup_perm_melt,
    )


def dedup_multiple_trimer(
        scenarios,
        methods,
        param_fpn,
):
    return DedupMultipleTrimer(
        scenarios=scenarios,
        methods=methods,
        param_fpn=param_fpn,
    )


def trace_single(
        df_apv,
        df_disapv,
):
    return TraceSingle(
        df_apv=df_apv,
        df_disapv=df_disapv,
    )


def trace_multiple(
        df_apv,
        df_disapv,
        scenarios,
        methods,
):
    return TraceMultiple(
        df_apv=df_apv,
        df_disapv=df_disapv,
        scenarios=scenarios,
        methods=methods,
    )


def inflat_exp(
        scenarios,
        methods,
        param_fpn,
):
    return Line(
        scenarios=scenarios,
        methods=methods,
        param_fpn=param_fpn,
    )


def prob_correct(
        error_rate: Union[float, np.ndarray]=0.00001,
        num_nt=12,
):
    TripletErrorCode(error_rate=error_rate).correct(num_nt=num_nt)


def prob_incorrect(
        error_rate: Union[float, np.ndarray]=0.00001,
        num_nt = 12,
):
    TripletErrorCode(error_rate=error_rate).incorrect(num_nt=num_nt)


def anchor_efficiency(
        quant_captured: Dict,
        quant_anchor_captured: Dict,
        criteria : List,
        condition : str,
):
    Anchor(
        quant_captured=quant_captured,
        quant_anchor_captured=quant_anchor_captured,
        criteria=criteria,
        condition=condition,
    ).draw()


def anchor_efficiency_broken(
        quant_captured: Dict,
        quant_anchor_captured: Dict,
        criteria : List,
        condition : str,
):
    Anchor(
        quant_captured=quant_captured,
        quant_anchor_captured=quant_anchor_captured,
        criteria=criteria,
        condition=condition,
    ).draw_line_broken()


def anchor_efficiency_simple(
        quant_captured: Dict,
        quant_anchor_captured: Dict,
        criteria : List,
        condition : str,
):
    Anchor(
        quant_captured=quant_captured,
        quant_anchor_captured=quant_anchor_captured,
        criteria=criteria,
        condition=condition,
    ).simple()


if __name__ == "__main__":
    from umiche.path import to

    plot_dm = dedup_multiple(
        scenarios={
            'pcr_nums': 'PCR cycle',
            'pcr_errs': 'PCR error',
            'seq_errs': 'Sequencing error',
            'ampl_rates': 'Amplification rate',
            'umi_lens': 'UMI length',
            'seq_deps': 'Sequencing depth',
        },
        methods={
            'unique': 'Unique',
            'cluster': 'Cluster',
            'adjacency': 'Adjacency',
            'directional': 'Directional',
            'dbscan_seq_onehot': 'DBSCAN',
            'birch_seq_onehot': 'Birch',
            'aprop_seq_onehot': 'Affinity Propagation',
            'mcl': 'MCL',
            'mcl_val': 'MCL-val',
            'mcl_ed': 'MCL-ed',
        },
        param_fpn=to('data/params.yml'),
    )
    # print(plot_dm.line())

    # print(graph_cluster(
    #     umi_uniq_val_cnt_fpn=to('data/simu/mclumi/seq_errs/umi_uniq_val_cnt-simu.txt'),
    #     # umi_uniq_val_cnt_fpn=to('data/simu/mclumi/seq_errs/umi_uniq_val_cnt-iclip.txt'),
    #     dedup_cluster=[
    #         [69, 72, 838, 1221, 97, 210, 249, 315, 324, 374, 457, 658, 727, 760, 771, 933, 1073, 1126, 1198, 1260, 1271,
    #          1307, 1498, 1505, 1541, 1563, 914, 946, 1083, 684, 1288, 1543, 822],
    #         [1174, 119, 290, 303, 204, 218, 289, 302, 404, 545, 586, 633, 674, 709, 720, 802, 884, 943, 980, 1355, 1436,
    #          1488, 1553, 786, 537, 867, 1649, 1255, 701, 1080, 347, 251], [1315, 1549]],
    #     ccs={69: [72, 838, 1221],
    #          72: [69, 97, 210, 249, 315, 324, 374, 457, 658, 727, 760, 771, 838, 933, 1073, 1126, 1198, 1221, 1260,
    #               1271, 1307, 1498, 1505, 1541, 1563], 838: [69, 72, 1221], 1221: [69, 72, 838, 914, 946],
    #          97: [72, 658, 1505], 210: [72, 1083, 1271], 249: [72, 684, 727, 1073], 315: [72, 374, 946, 1126],
    #          324: [72, 771], 374: [72, 315, 1126], 457: [72, 1288, 1307], 658: [72, 97, 1505], 727: [72, 249, 1073],
    #          760: [72, 933, 1543], 771: [72, 324], 933: [72, 760], 1073: [72, 249, 727], 1126: [72, 315, 374],
    #          1198: [72, 1260, 1541], 1260: [72, 1083, 1198, 1541, 1543], 1271: [72, 210], 1307: [72, 457],
    #          1498: [72, 684, 1563], 1505: [72, 97, 658, 822], 1541: [72, 1198, 1260], 1563: [72, 914, 1498],
    #          914: [1221, 1563], 946: [315, 1221], 1083: [210, 1260], 684: [249, 1498], 1288: [457],
    #          1543: [760, 1174, 1260], 822: [1505], 1174: [119, 290, 303, 1315, 1543],
    #          119: [204, 218, 289, 290, 302, 303, 404, 545, 586, 633, 674, 709, 720, 802, 884, 943, 980, 1174, 1355,
    #                1436, 1488, 1553], 290: [119, 303, 786, 1174], 303: [119, 290, 1174], 1315: [1174, 1549],
    #          204: [119, 537, 720, 1488], 218: [119, 302, 674, 867, 1649], 289: [119, 980],
    #          302: [119, 218, 674, 1255], 404: [119, 802, 943], 545: [119, 701, 1436], 586: [119, 709, 884],
    #          633: [119, 1080, 1553], 674: [119, 218, 302], 709: [119, 586, 884], 720: [119, 204, 347, 1488],
    #          802: [119, 404, 537, 943], 884: [119, 586, 701, 709], 943: [119, 404, 802], 980: [119, 251, 289],
    #          1355: [119, 251, 867], 1436: [119, 545], 1488: [119, 204, 720], 1553: [119, 633, 786], 786: [290, 1553],
    #          1549: [1315], 537: [204, 802], 867: [218, 1355, 1649], 1649: [218, 867], 1255: [302], 701: [545, 884],
    #          1080: [633], 347: [720], 251: [980, 1355]},
    #     edge_list=[(97, 72), (658, 97), (1355, 119), (204, 119), (1649, 218), (1541, 72), (933, 72), (914, 1563),
    #                (1505, 72), (1221, 69), (404, 119), (1649, 867), (1307, 72), (302, 119), (802, 119), (1543, 760),
    #                (1126, 72), (709, 586), (786, 1553), (374, 315), (946, 315), (1073, 72), (720, 204), (933, 760),
    #                (1174, 1543), (210, 72), (771, 72), (537, 204), (727, 249), (1563, 72), (684, 1498), (822, 1505),
    #                (1549, 1315), (1505, 97), (545, 119), (374, 72), (1553, 119), (1436, 119), (867, 218), (457, 72),
    #                (1563, 1498), (709, 119), (720, 119), (1260, 1198), (1543, 1260), (218, 119), (251, 1355),
    #                (701, 545), (1541, 1260), (1488, 204), (1553, 633), (1083, 210), (980, 289), (119, 1174),
    #                (727, 72), (290, 119), (1073, 727), (1541, 1198), (884, 586), (802, 404), (303, 119), (347, 720),
    #                (674, 218), (1221, 838), (1221, 72), (914, 1221), (943, 119), (586, 119), (943, 802), (838, 72),
    #                (1271, 210), (980, 119), (674, 302), (786, 290), (251, 980), (1488, 720), (1488, 119),
    #                (946, 1221), (1073, 249), (884, 709), (1288, 457), (1126, 315), (943, 404), (537, 802),
    #                (302, 218), (249, 72), (315, 72), (303, 290), (290, 1174), (884, 119), (633, 119), (771, 324),
    #                (658, 72), (1505, 658), (1307, 457), (303, 1174), (72, 69), (289, 119), (838, 69), (684, 249),
    #                (1498, 72), (1315, 1174), (1126, 374), (701, 884), (1255, 302), (674, 119), (1080, 633),
    #                (1083, 1260), (1436, 545), (1260, 72), (1271, 72), (867, 1355), (760, 72), (1198, 72), (324, 72)],
    #
    # ).draw())

    # print(inflat_exp(
    #     scenarios={
    #         'pcr_nums': 'PCR cycle',
    #         'pcr_errs': 'PCR error',
    #         'seq_errs': 'Sequencing error',
    #         'ampl_rates': 'Amplification rate',
    #         'umi_lens': 'UMI length',
    #         'seq_deps': 'Sequencing depth',
    #     },
    #     methods={
    #         'mcl': 'MCL',
    #     },
    #     param_fpn=to('data/params.yml'),
    # ).draw())

    # from umiche.io import stat
    # scenarios = {
    #     # 'pcr_nums': 'PCR cycle',
    #     # 'pcr_errs': 'PCR error',
    #     'seq_errs': 'Sequencing error',
    #     # 'ampl_rates': 'Amplification rate',
    #     # 'umi_lens': 'UMI length',
    #     # 'seq_deps': 'Sequencing depth',
    # }
    # methods = {
    #     # 'unique': 'Unique',
    #     # 'cluster': 'Cluster',
    #     # 'adjacency': 'Adjacency',
    #     'directional': 'Directional',
    #     # 'dbscan_seq_onehot': 'DBSCAN',
    #     # 'birch_seq_onehot': 'Birch',
    #     # 'aprop_seq_onehot': 'Affinity Propagation',
    #     'mcl': 'MCL',
    #     'mcl_val': 'MCL-val',
    #     'mcl_ed': 'MCL-ed',
    # }
    # dedupstat = stat(
    #     scenarios=scenarios,
    #     methods=methods,
    #     param_fpn=to('data/params.yml'),
    # )
    # df_dedup = dedupstat.df_dedup
    # df_dedup_perm_melt = dedupstat.df_dedup_perm_melt
    # t = dedup_single(
    #     df_dedup=df_dedup,
    #     df_dedup_perm_melt=df_dedup_perm_melt,
    # )
    # # print(t.jointgrid(
    # #     # method='Directional',
    # #     # method='MCL',
    # #     # method='MCL-val',
    # #     method='MCL-ed',
    # # ))
    # # print(t.stackedbar())
    # # print(t.strip())

    # print(dedup_multiple_trimer(
    #     scenarios={
    #         'seq_errs': 'Sequencing error rate',
    #     },
    #     methods={
    #         'directional_dedupby_majority_vote_splitby__collblockby_take_by_order': 'UMI-tools+drMV+cbRAN',
    #         'directional_dedupby_majority_vote_splitby__collblockby_majority_vote': 'UMI-tools+drMV+cbMV',
    #         'directional_dedupby_set_cover_splitby_split_by_mv_collblockby_take_by_order': 'UMI-tools+drSC+spMV+cbRAN',
    #         'directional_dedupby_set_cover_splitby_split_by_mv_collblockby_majority_vote': 'UMI-tools+drSC+spMV+cbMV',
    #         'directional_dedupby_set_cover_splitby_split_to_all_collblockby_take_by_order': 'UMI-tools+drSC+spALL+cbRAN',
    #         'directional_dedupby_set_cover_splitby_split_to_all_collblockby_majority_vote': 'UMI-tools+drSC+spALL+cbMV',
    #     },
    #     param_fpn=to('data/params_trimer.yml'),
    # ).line())


    ### ++++++++++++++++++++trace++++++++++++++++++++++++++
    from umiche.deduplicate.io.Stat import Stat as dedupstat
    scenarios = {
        'pcr_nums': 'PCR cycle',
        # 'pcr_errs': 'PCR error',
        # 'seq_errs': 'Sequencing error',
        # 'ampl_rates': 'Amplification rate',
        # 'umi_lens': 'UMI length',
        # 'seq_deps': 'Sequencing depth',
    }
    methods = {
        # 'unique': 'Unique',
        # 'cluster': 'Cluster',
        # 'adjacency': 'Adjacency',
        'directional': 'Directional',
        # 'dbscan_seq_onehot': 'DBSCAN',
        # 'birch_seq_onehot': 'Birch',
        # 'aprop_seq_onehot': 'Affinity Propagation',
        # 'mcl': 'MCL',
        # 'mcl_val': 'MCL-val',
        # 'mcl_ed': 'MCL-ed',
    }
    dedupstat11 = dedupstat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=to('data/params.yml'),
    )
    # trace_single(
    #     df_apv=dedupstat11.df_trace_cnt['apv'],
    #     df_disapv=dedupstat11.df_trace_cnt['disapv'],
    # ).line_apv_disapv()

    scenarios = {
        'pcr_nums': 'PCR cycle',
        # 'pcr_errs': 'PCR error',
        # 'seq_errs': 'Sequencing error',
        # 'ampl_rates': 'Amplification rate',
        # 'umi_lens': 'UMI length',
        # 'seq_deps': 'Sequencing depth',
    }
    methods = {
        'unique': 'Unique',
        'cluster': 'Cluster',
        'adjacency': 'Adjacency',
        'directional': 'Directional',
        'dbscan_seq_onehot': 'DBSCAN',
        'birch_seq_onehot': 'Birch',
        'aprop_seq_onehot': 'Affinity Propagation',
        'mcl': 'MCL',
        'mcl_val': 'MCL-val',
        'mcl_ed': 'MCL-ed',
    }
    dedupstat22 = dedupstat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=to('data/params.yml'),
    )

    # trace_multiple(
    #     df_apv=dedupstat22.df_trace_cnt['apv'],
    #     df_disapv=dedupstat22.df_trace_cnt['disapv'],
    #     scenarios=scenarios,
    #     methods=methods,
    # ).line_apv()

    error_rates= np.linspace(0.00001, 0.5, 500)
    prob_correct(error_rate=error_rates, num_nt=1)
    prob_incorrect(error_rate=error_rates, num_nt=12)


import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.util.Reader import Reader as freader
from umiche.deduplicate.trimer.pipeline import Config


class protDTSplit(Config.config):

    def __init__(self, fpns):
        super(protDTSplit, self).__init__()
        self.freader = freader()
        self.df = pd.DataFrame()
        self.df_T = pd.DataFrame()
        for method, fpn in fpns.items():
            df_met = self.freader.generic(df_fpn=fpn, header=0)[:]
            # df_met = self.freader.generic(df_fpn=fpn, header=0)[:13]
            df_met_T = df_met.T
            df_met_T = (df_met_T - 50) / 50
            df_met_T.columns = ['{:.1e}'.format(x) for x in self.seq_errs[:]]
            # df_met_T.columns = [int(x*100000) for x in self.seq_errs[:]]
            df_met_T['method'] = method
            # print(df_met_T)
            # df_met = np.exp((df_met - 50) / 50)
            df_met = (df_met - 50) / 50
            df_met['mean'] = df_met.mean(axis=1)
            df_met['max'] = df_met.max(axis=1)
            df_met['min'] = df_met.min(axis=1)
            df_met['std'] = df_met.std(axis=1)
            df_met['mean-min'] = df_met['std']
            df_met['max-mean'] = df_met['std']
            df_met['method'] = method
            df_met['metric'] = ['{:.1e}'.format(x) for x in self.seq_errs[:]]
            self.df = pd.concat([self.df, df_met], axis=0)
            self.df_T = pd.concat([self.df_T, df_met_T], axis=0)
        print(self.df)
        self.df_direc = self.df[self.df['method'] == 'directional']
        self.df_mcl_val = self.df[self.df['method'] == 'mcl_val']
        self.df_mcl_ed = self.df[self.df['method'] == 'mcl_ed']
        # print(self.df_T)
        # self.df = self.df.reset_index(drop=True)

        # sns.jointplot(data=self.df, x="mean-min", y="max", hue="met",)
        plt.show()
        self.df_melt = pd.melt(self.df_T, 'method', var_name="Sequencing error")
        # self.df_melt = pd.DataFrame(self.df_melt)
        print(self.df_melt)
        self.df_gp = self.df.groupby(by=['method'])
        self.gp_keys = self.df_gp.groups.keys()
        # print(self.gp_keys)

    def jointplot(self, ):
        # fig, ax = plt.subplots()
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        ppp = sns.jointplot(
            # x=self.df_mcl_ed['mean'].values,
            x=self.df_mcl_val['mean'].values,
            y=self.df_direc['mean'].values,
            kind="reg",
            color="crimson",
            label='asd',
        )
        ppp.ax_joint.plot([0, 50], [0, 50], 'grey', linewidth=2, alpha=1)
        # ppp.set_axis_labels('mcl_ed '+ r'($\frac{N_e-N_t}{N_t}$)' , 'directional ' + r'($\frac{N_e-N_t}{N_t}$)', fontsize=14)
        ppp.set_axis_labels('mcl_val '+ r'($\frac{N_e-N_t}{N_t}$)' , 'directional ' + r'($\frac{N_e-N_t}{N_t}$)', fontsize=14)
        ppp.ax_joint.text(40, 45, "confidence interval", horizontalalignment='right', size='medium', color='crimson',)
        ppp.ax_joint.text(20, 35, "regression", horizontalalignment='left', size='medium', color='crimson',)
        ppp.ax_joint.text(34, 32, "baseline", horizontalalignment='left', size='medium', color='black',)
        sns.despine(right=True, top=True)
        plt.tight_layout()
        plt.show()

    def jointgrid(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        # fig, ax = plt.subplots()
        # met = 'directional'
        # met = 'mcl_ed'
        met = 'mcl_val'

        ddd = pd.DataFrame()
        ddd['method'] = self.df_melt['method']
        ddd['Sequencing error'] = self.df_melt['Sequencing error']
        ddd[r'$\frac{N_e-N_t}{N_t}$'] = self.df_melt['value']

        g = sns.JointGrid(data=ddd[ddd['method'] == met], x="Sequencing error", y=r'$\frac{N_e-N_t}{N_t}$', marginal_ticks=True)

        # Create an inset legend for the histogram colorbar
        cax = g.figure.add_axes([.20, .55, .02, .2])

        # Add the joint and marginal histogram plots
        g.plot_joint(
            sns.kdeplot, discrete=(True, False),
            cmap="light:#03012d", pmax=.8, cbar=True, cbar_ax=cax
        )

        g.ax_joint.set_title(met, fontsize=14)
        g.ax_joint.set_xticks([int(x*100000) for x in self.seq_fix_errs[:]])
        g.ax_joint.set_xticklabels([1e-05, '', '', '', '', '', '', '', 0.001, 0.0025, 0.005, 0.0075, 0.01])
        plt.setp(g.ax_joint.get_xticklabels(), rotation=45)

        g.plot_marginals(
            sns.histplot,
            element="step",
            color="#03012d",
        )
        # ax.legend(ncol=2, loc="upper right", frameon=True)
        # g.set(ylabel="", xlabel="Automobile collisions per billion miles")
        # ax.set_ylabel('Sequencing error', fontsize=12)
        # ax.set_xlabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=12)
        # sns.despine(left=True, bottom=True)
        plt.tight_layout()
        plt.show()
        return

    def strip(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        fig, ax = plt.subplots()
        sns.despine(bottom=True, left=True)
        cc = [
            'tab:green',
            'crimson',
            'tab:blue',
        ]
        # Show each observation with a scatterplot
        sns.stripplot(x="value", y="Sequencing error", hue='method', palette=cc,
                      data=self.df_melt, dodge=True, alpha=.25, zorder=1)
        sns.pointplot(x="value", y="Sequencing error", hue='method',
                      data=self.df_melt, dodge=.8 - .8 / 3,
                      join=False,  palette=cc,
                      markers="d", scale=.75, ci=None)

        # Improve the legend
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[3:], labels[3:], title='method',
                  handletextpad=0, columnspacing=1,
                  loc="upper right", ncol=3, frameon=True)
        ax.set_xlabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=14)
        ax.set_ylabel('Sequencing error rate', fontsize=14)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.15,
            left=0.15,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.show()

    def stackedbar(self, ):
        fig, ax = plt.subplots(figsize=(4, 5))
        # sns.set(font="Verdana")
        sns.set(font="Helvetica")

        self.df_mcl_val["dmean"] = self.df_direc["mean"] - self.df_mcl_val["mean"]
        self.df_mcl_ed["dmean"] = self.df_direc["mean"] - self.df_mcl_ed["mean"]

        # self.df_mcl_val["dmean"] = np.exp(self.df_direc["mean"] - self.df_mcl_val["mean"])
        # self.df_mcl_ed["dmean"] = np.exp(self.df_direc["mean"] - self.df_mcl_ed["mean"])

        sns.set_color_codes("pastel")
        sns.barplot(
            x="dmean",
            y="metric",
            data=self.df_mcl_val,
            label="dFC_ed",
            color="b",
        )
        sns.set_color_codes("muted")
        sns.barplot(
            x="dmean",
            y="metric",
            data=self.df_mcl_ed,
            label="dFC_val",
            color="b",
        )

        # plt.fill_between(self.df_mcl_val["dmean"], self.df_mcl_val["metric"])
        ax.legend(ncol=2, loc="upper right", frameon=True)
        # ax.set(ylabel="", xlabel="Automobile collisions per billion miles")
        ax.set_ylabel('Sequencing error', fontsize=12)
        ax.set_xlabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=12)
        sns.despine(left=True, bottom=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.12,
            left=0.20,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.show()

    def errorbar(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        # plt.rc('text', usetex=True)
        # plt.rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
        fig, ax = plt.subplots()
        # ax.plot(
        #     file_n301.index,
        #     n301_means,
        #     label='Full TrainData dataset',
        #     alpha=0.9,
        #     linewidth=3.0,
        #     # s=,
        #     c='royalblue'
        # )
        cc = [
            'tab:green',
            'tab:blue',
            'crimson',
            'tab:orange',
            'tab:brown',
        ]

        for i, met in enumerate(self.gp_keys):
            df_met = self.df_gp.get_group(met)

            ax.errorbar(
                x=df_met.index,
                y=df_met['mean'],
                yerr=[df_met['mean-min'], df_met['max-mean']],
                fmt='o',
                alpha=0.6,
                ecolor=cc[i],
                color=cc[i],
                linewidth=1,
                elinewidth=1,
                capsize=2,
                markersize=6,
                label=met,
            )
            # print(df_met['metric'])
            ax.set_xticks(np.arange(df_met['metric'].shape[0]))
            ax.set_xticklabels(df_met['metric'], fontsize=8, rotation=45)
        # ax.set_xlabel('Sequencing error rate', fontsize=12)
        # ax.set_xlabel('Amplification rate', fontsize=12)
        # ax.set_xlabel('UMI length', fontsize=12)
        ax.set_xlabel('Polymerase error rate', fontsize=12)
        ax.set_ylabel(r'$\frac{N_e-N_t}{N_t}$', fontsize=16)
        # ax.set_title(DEFINE['title'], fontsize=12)
        sns.despine(right=True, top=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.16,
            left=0.11,
            # left=0.12,
            right=0.95,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.legend(fontsize=11, loc='upper left')
        plt.show()
        return

    def errorband(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        fig, ax = plt.subplots()
        cc = [
            'tab:green',
            # 'black',
            'tab:blue',
            # 'tab:orange',
            'midnightblue',
            'crimson',

            'tab:brown',

            'cornflowerblue',
            'maroon',
            'lightcoral',
        ]
        for i, met in enumerate(self.gp_keys):
            df_met = self.df_gp.get_group(met)
            ax.errorbar(
                x=df_met.index,
                y=df_met['mean'],
                yerr=[df_met['mean-min'], df_met['max-mean']],
                fmt='o',
                alpha=0.7,
                ecolor=cc[i],
                color=cc[i],
                linestyle='-',
                linewidth=2,
                elinewidth=0.5,
                capsize=2,
                markersize=3,
                label=met,
            )
            ax.plot(df_met.index, df_met['mean'] - df_met['mean-min'], color=cc[i], linewidth=0.1,alpha=0.1)
            ax.plot(df_met.index, df_met['max-mean'] + df_met['mean'], color=cc[i], linewidth=0.1,alpha=0.1)
            ax.fill_between(
                df_met.index,
                df_met['mean'] - df_met['mean-min'],
                df_met['max-mean'] + df_met['mean'],
                alpha=0.1,
                color=cc[i],
            )
            ax.set_xticks(df_met['metric'].index)
            ax.set_xticklabels(df_met['metric'], fontsize=8, rotation=45)

        ax.set_xlabel('Sequencing error rate', fontsize=12)
        # ax.set_xlabel('Amplification rate', fontsize=12)
        ax.set_ylabel('Fold change (' + r'$\frac{N_e-N_t}{N_t}$' + ')', fontsize=14)
        # ax.set_title(DEFINE['title'], fontsize=12)
        sns.despine(right=True, top=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.16,
            left=0.13,
            right=0.95,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.legend(fontsize=11)
        plt.show()
        return

    def errbar(self, ):
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        fig, ax = plt.subplots()
        # fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(8, 3), sharey=False, sharex='all')
        cc = [
            # 'tab:green',
            # 'black',
            # 'tab:blue',
            # 'tab:orange',
            'seagreen',
            'pink',
            'midnightblue',
            'crimson',

            'tab:brown',

            'cornflowerblue',
            'maroon',
            'lightcoral',
        ]

        for i, met in enumerate(self.gp_keys):
            df_met = self.df_gp.get_group(met)
            print(df_met.iloc[-9:,:])
            ax.errorbar(
                x=df_met.iloc[-9:,:].index - 0.2 if i == 1 else df_met.iloc[-9:,:].index + 0.2,
                y=df_met.iloc[-9:,:]['mean'],
                yerr=[df_met.iloc[-9:,:]['mean-min'], df_met.iloc[-9:,:]['max-mean']],
                fmt='o',
                alpha=0.7,
                ecolor=cc[i],
                # color='white',
                # linestyle='-',
                # linewidth=2,
                elinewidth=2,
                capsize=3,
                markersize=0.5,
                # label=met,
            )
            ax.bar(
                df_met.iloc[-9:,:].index - 0.20 if i == 1 else df_met.iloc[-9:,:].index + 0.20,
                height=df_met.iloc[-9:,:]['mean'],
                width=0.4,
                color=cc[i],  # 'gainsboro'
                label=met,
                alpha=0.9,
                # edgecolor='black',
                linewidth=0.001,
            )
            # ax.plot(df_met.iloc[-9:,:].index, df_met.iloc[-9:,:]['mean'] - df_met.iloc[-9:,:]['mean-min'], color=cc[i], linewidth=0.1, alpha=0.1)
            # ax.plot(df_met.iloc[-9:,:].index, df_met.iloc[-9:,:]['max-mean'] + df_met.iloc[-9:,:]['mean'], color=cc[i], linewidth=0.1, alpha=0.1)
            # ax.fill_between(
            #     df_met.iloc[-9:,:].index,
            #     df_met.iloc[-9:,:]['mean'] - df_met.iloc[-9:,:]['mean-min'],
            #     df_met.iloc[-9:,:]['max-mean'] + df_met.iloc[-9:,:]['mean'],
            #     alpha=0.1,
            #     color=cc[i],
            # )
            ax.set_xticks(df_met.iloc[-9:,:]['metric'].index)
            ax.set_xticklabels(df_met.iloc[-9:,:]['metric'], fontsize=10, rotation=30)

        ax.set_xlabel('Sequencing error rate', fontsize=12)
        # ax.set_xlabel('Amplification rate', fontsize=12)
        ax.set_ylabel('Fold change (' + r'$\frac{N_e-N_t}{N_t}$' + ')', fontsize=14)
        # ax.set_title(DEFINE['title'], fontsize=12)
        sns.despine(right=True, top=True)
        fig.subplots_adjust(
            top=0.98,
            bottom=0.16,
            left=0.12,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15
        )
        plt.legend(fontsize=11)
        plt.show()
        return


if __name__ == "__main__":
    from umiche.path import to
    # metric_char ='pcr_nums'
    # metric_char ='pcr_errs'
    metric_char ='seq_errs'
    # metric_char  = 'ampl_rates'
    # metric_char ='umi_lens'

    DEFINE = {
        'fpns': {
            # # 'directional_mono': to('data/simu/trimer/pcr8/') + metric_char + '/directional_ref.txt',
            # 'directional_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/directional_bipartite.txt',
            # 'mcl_val_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_bipartite.txt',
            # 'mcl_ed_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_ed_bipartite.txt',
            # # 'directional_dimer (no correction)': to('data/simu/dimer/pcr8/') + metric_char + '/directional_bipartite.txt',

            # 'monomer12-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_ref.txt',
            # # 'directional_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/directional_bipartite.txt',
            # 'trimer-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_bipartite.txt',
            # # 'mcl_ed_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_ed_bipartite.txt',
            # 'monomer12-NC': to('data/simu/trimer/pcr8/') + metric_char + '/unique_ref.txt',
            # 'dimer-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_ed_bipartite_di.txt',
            # 'monomer24-NC': to('data/simu/trimer/pcr8/') + metric_char + '/unique_24.txt',
            # 'monomer24-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_24.txt',
            # 'monomer36-NC': to('data/simu/trimer/pcr8/') + metric_char + '/unique_36.txt',
            # 'monomer36-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_36.txt',
            # # # 'directional_mono': to('data/simu/trimer/pcr8/') + metric_char + '/directional_ref.txt',
            # # 'directional trimer': to('data/simu/trimer/pcr8/') + metric_char + '/directional_bipartite.txt',
            # # 'mcl trimer': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_bipartite.txt',
            # # # 'mcl_ed_trimer': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_ed_bipartite.txt',
            # # # 'monomer - no correction': to('data/simu/trimer/pcr8/') + metric_char + '/unique_ref.txt',
            # # # 'dimer - corrected': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_ed_bipartite_di.txt',


            # 'UMI-tools': to('data/simu/trimer/pcr8/') + metric_char + '/directional_bipartite.txt',
            # 'mclUMI': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_bipartite.txt',


            # 'monomer-NC': to('data/simu/trimer/pcr8/') + metric_char + '/unique_lmr.txt',
            # 'trimer-MC': to('data/simu/trimer/pcr8/') + metric_char + '/unique_bipartite.txt',
            # 'monomer-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_lmr.txt',
            # 'trimer-MC-Mclumi': to('data/simu/trimer/pcr8/') + metric_char + '/mcl_val_bipartite.txt',
            # # 'monomer - ref': to('data/simu/trimer/pcr8/') + metric_char + '/unique_ref.txt',

            # 'monomer': to('data/simu/trimer/pcr8/') + metric_char + '/unique_lmr.txt',
            # 'trimer majority vote': to('data/simu/trimer/pcr8/') + metric_char + '/unique_bipartite.txt',
            # 'trimer set cover': to('data/simu/trimer/pcr8/') + metric_char + '/set_cover_trimmed.txt',

            'monomer - without correction': to('data/simu/trimer/pcr8/') + metric_char + '/unique_mononer.txt',
            'monomer - UMI-tools (directional)': to('data/simu/trimer/pcr8/') + metric_char + '/directional_mononer.txt',
            'trimer - majority vote': to('data/simu/trimer/pcr8/') + metric_char + '/unique_bipartite.txt',
            'trimer - set cover': to('data/simu/trimer/pcr8/') + metric_char + '/set_cover_trimmed.txt',

        },
    }
    p = protDTSplit(
        fpns=DEFINE['fpns']
    )
    # print(p.strip())
    # print(p.jointplot())
    # print(p.jointgrid())
    # print(p.stackedbar())
    # print(p.errorbar())
    print(p.errorband())
    # print(p.errbar())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"


import tensorflow as tf
from tensorflow.keras import backend as K


class condiConvVAE:

    def __init__(
            self,
            input_shape,
            image_size,
            label_shape,
            batch_size,
            kernel_size,
            filters,
            latent_dim,
            strides,
            epochs,
            inputs,
            y_labels,
    ):
        self.image_size = image_size
        self.input_shape = input_shape
        self.label_shape = label_shape
        self.batch_size = batch_size
        self.kernel_size = kernel_size
        self.filters = filters
        self.latent_dim = latent_dim
        self.strides = strides
        self.epochs = epochs
        self.inputs = inputs
        self.y_labels = y_labels

    def reparameterize(self, params):
        z_mean, z_log_var = params
        batch = K.shape(z_mean)[0]
        dim = K.int_shape(z_mean)[1]
        epsilon = K.random_normal(shape=(batch, dim))
        return z_mean + K.exp(0.5 * z_log_var) * epsilon

    def loss(self, outputs, z_mean, z_log_var):
        # VAE loss = mse_loss or xent_loss + kl_loss
        beta = 1.0
        # rloss = tf.keras.losses.binary_crossentropy(K.flatten(inputs), K.flatten(outputs))
        rloss = tf.keras.losses.mse(K.flatten(self.inputs), K.flatten(outputs))
        rloss *= self.image_size * self.image_size
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = K.sum(kl_loss, axis=-1)
        kl_loss *= -0.5 * beta
        return K.mean(rloss + kl_loss)

    def encoding(self, ):
        x = tf.keras.layers.Dense(self.image_size * self.image_size)(self.y_labels)
        x = tf.keras.layers.Reshape((self.image_size, self.image_size, 1))(x)
        x = tf.keras.layers.concatenate([self.inputs, x])
        x = tf.keras.layers.Conv2D(self.filters, self.kernel_size, self.strides, padding='same', activation='relu')(x)
        x = tf.keras.layers.Conv2D(self.filters * 2, self.kernel_size, self.strides, padding='same', activation='relu')(x)
        # for decoder
        shape = K.int_shape(x)
        # generate latent vector Q(z|X)
        x = tf.keras.layers.Flatten()(x)
        x = tf.keras.layers.Dense(16, activation='relu')(x)
        z_mean = tf.keras.layers.Dense(self.latent_dim, name='z_mean')(x)
        z_log_var = tf.keras.layers.Dense(self.latent_dim, name='z_log_var')(x)

        # use reparameterization trick to push the sampling out as input
        z = tf.keras.layers.Lambda(self.reparameterize, output_shape=(self.latent_dim,), name='z',)([z_mean, z_log_var])
        encoder = tf.keras.models.Model(inputs=[self.inputs, self.y_labels], outputs=[z_mean, z_log_var, z], name='encoder')
        return shape, z_mean, z_log_var, encoder

    def decoding(self, shape,):
        latent_inputs = tf.keras.layers.Input(shape=(self.latent_dim,), name='z_sampling')
        x = tf.keras.layers.concatenate([latent_inputs, self.y_labels])
        x = tf.keras.layers.Dense(shape[1]*shape[2]*shape[3], activation='relu')(x)
        x = tf.keras.layers.Reshape((shape[1], shape[2], shape[3]))(x)
        x = tf.keras.layers.Conv2DTranspose(self.filters, self.kernel_size, self.strides, padding='same', activation='relu')(x)
        x = tf.keras.layers.Conv2DTranspose(self.filters / 2, self.kernel_size, self.strides, padding='same', activation='relu')(x)
        x = tf.keras.layers.Conv2DTranspose(1, self.kernel_size, padding='same', activation='sigmoid', name='decoder_output')(x)
        decoder = tf.keras.models.Model(inputs=[latent_inputs, self.y_labels], outputs=x, name='decoder')
        return decoder


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
import umiche.plot.bar.external.lca_standard_graphs as lsg
from umiche.util.Reader import Reader as freader
from umiche.path import to


class editdistance:

    def __init__(self, ):
        self.freader = freader()

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def umi_tools(self, ):
        tool = 'UMI-tools'
        df_raw = self.freader.generic(
            df_fpn=to('ext/data/seq/umi/example.bam/ed1/umi-tools/') + 'deduplicated_edit_distance.tsv',
            df_sep='\t',
            header=0,
            skiprows=1,
        )
        df_raw.columns = ['unique', 'unique_null', 'direc', 'direc_null', 'ed']
        unique = df_raw[['unique', 'ed']]
        unique = unique.rename(columns={'unique': 'cnt'}).T
        unique['met'] = 'unique'
        unique['tool'] = tool
        print(unique)

        direc = df_raw[['direc', 'ed']]
        direc = direc.rename(columns={'direc': 'cnt'}).T
        direc['met'] = 'directional'
        direc['tool'] = tool
        print(direc)
        df = pd.concat([unique, direc])
        return df

    def data(self, ):
        df_umi_tools = self.umi_tools()
        tool = 'Mclumi'
        umi_len = 12
        met_full = ['unique', 'cluster', 'adjacency', 'directional', 'mcl', 'mcl_val', 'mcl_ed']
        met_abbr = ['uniq', 'cc', 'adj', 'direc', 'mcl', 'mcl_val', 'mcl_ed']
        df = pd.DataFrame()
        for met_a, met_f in zip(met_abbr, met_full):
            asd = self.freader.generic(
                # df_fpn=to('ext/data/seq/umi/example.bam/ed1/mclumi/') + met_a + '_ave_ed_pos_bin.txt',
                df_fpn=to('ext/data/bulk/RM82CLK1_S3/umi/ed7/') + met_a + '_ave_ed_pos_bin.txt',
                # df_fpn=to('ext/data/sc/umi/hgmm100/ed6/') + met_a + '_ave_ed_pos_bin.txt',
                df_sep='\t',
                header=None,
                skiprows=1,
            )
            asd.columns = ['ed', 'cnt']
            t = dict(zip(asd.ed, asd.cnt))
            for i in range(umi_len + 1):
                if i not in [*t.keys()]:
                    t[i] = 0
            c = pd.DataFrame({'ed': t.keys(), 'cnt': t.values()}).sort_values(by='ed').reset_index(drop=True)
            c = c.T
            c['met'] = met_f
            c['tool'] = tool
            # print(c)
            df = pd.concat([df, c])
            print(df)
        # df = pd.concat([df, df_umi_tools])
        print(df)
        df = df.loc[df.index == 'cnt']
        # df = df.loc[(df.met == 'unique') | (df.met == 'directional')]
        df = df.rename(columns={i: 'ave edit distance ' + str(i) for i in range(umi_len + 1)})
        df1 = df.set_index(['met', 'tool'])
        print(df1)

        ax, fig = lsg.plot_grouped_stackedbars(
            df1, ix_categories='met',
            ix_entities_compared='tool',

            # df1, ix_categories='tool',
            # ix_entities_compared='met',

            xaxis_label='Relative (%) number of genomic positions observed',
            figsize=(10, 4)
        )
        plt.show()
        return

    def wk2(self, ):
        tool = 'Mclumi'
        umi_len = 10
        met_full = ['Cluster', 'Adjacency', 'Directional', 'MCL']
        met_full_s = ['cluster', 'adjacency', 'directional', 'mcl']
        met_abbr = ['cc', 'adj', 'direc', 'mcl']
        # df = pd.DataFrame()
        df_ex = self.freader.generic(
            df_fpn=to('data/simu/sc/hgmm100/ed6/uniq_dedup_sum.txt'),
            df_sep='\t',
            header=0,
            index_col=0,
            # skiprows=1,
        )
        df_ex['cell'] = df_ex.apply(lambda x: x.name.split(', ')[0].split('(')[1].split("'")[1], axis=1)
        df_ex['gene'] = df_ex.apply(lambda x: x.name.split(', ')[1].split(')')[0].split("'")[1], axis=1)
        cell_types = df_ex['cell'].unique()
        cell_maps = {v: i for i, v in enumerate(cell_types)}
        gene_types = df_ex['gene'].unique()
        gene_maps = {v: i for i, v in enumerate(gene_types)}
        gene_rev_maps = {i: v for i, v in enumerate(gene_types)}
        for met_a, met_f in zip(met_abbr, met_full):
            df = self.freader.generic(
                # df_fpn=to('ext/data/seq/umi/example.bam/ed1/mclumi/') + met_a + '_ave_ed_pos_bin.txt',
                # df_fpn=to('ext/data/bulk/RM82CLK1_S3/umi/') + met_a + '_ave_ed_pos_bin.txt',
                df_fpn=to('data/simu/sc/hgmm100/ed6/') + met_a + '_dedup_sum.txt',
                df_sep='\t',
                header=0,
                index_col=0,
                # skiprows=1,
            )
            df['cell'] = df.apply(lambda x: x.name.split(', ')[0].split('(')[1].split("'")[1], axis=1)
            df['gene'] = df.apply(lambda x: x.name.split(', ')[1].split(')')[0].split("'")[1], axis=1)
            df['cell_type'] = df['cell'].apply(lambda x: cell_maps[x])
            df['gene_type'] = df['gene'].apply(lambda x: gene_maps[x])
            # print(df.loc[df['cell_type'].isin(range(20))])
            # print(df.loc[df['cell_type'].isin(range(20))])
            from scipy.sparse import csr_matrix
            gc_mat = csr_matrix(
                (df[met_a + '_umi_len'].values, (df['cell_type'].values, df['gene_type'].values)),
                shape=(len(cell_maps), len(gene_maps))
            )
            print('===>gc matrix shape: {}'.format(gc_mat.shape))
            from umiche.plot.stat.gene.expression.Extreme import Extreme as ppp
            high_genes = ppp().geneIds(
                gc_mat=gc_mat.toarray(),
                oriented='maximal',
                n_top=10,
                scheme={'method': 'mean', 'cell_ratio': 0.9},
            )
            gene_ids = [gene_rev_maps[i] for i in high_genes]
            print(gene_ids)
            newdfd = df.loc[df['gene'].isin(gene_ids)]
            self.bplot(x_col_name=met_a + '_umi_len', y_col_name='gene', df=newdfd, xlabel=met_f)

            # for i in gene_ids:
            #     # print(df.loc[df['gene'].isin([gene_rev_maps[i]])][met_a + '_umi_len'].values)
            #     print(df.loc[df['gene'].isin([i])])
        return

    def asdas(self, ):
        import numpy as np
        np.random.seed(19680801)
        data = np.random.lognormal(size=(37, 4), mean=1.5, sigma=1.75)
        labels = list('ABCD')
        fs = 10  # fontsize
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 6), sharey=True)
        bplot = ax.boxplot(data, labels=labels)
        ax.set_title('Default', fontsize=fs)
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')

        # for bplot_handle, color in zip(boxplot_handles, palette):
        for patch in bplot['boxes']:
            patch.set(color='black', linewidth=1.1)
            # patch.set(facecolor='white', alpha=0.3)
        for whisker in bplot['whiskers']:
            whisker.set(color='black', linewidth=1)
        for cap in bplot['caps']:
            cap.set(color='black', linewidth=1)
        for median in bplot['medians']:
            median.set(color='black', linewidth=1)
        for flier in bplot['fliers']:
            flier.set(marker='o', color='y', alpha=0.5)

        fig.subplots_adjust(hspace=0.4)
        plt.show()

    def bplot(self, x_col_name, y_col_name, xlabel, df):
        fig, ax = plt.subplots(figsize=(8, 6))
        # ax.set_xscale("log")

        # Plot the orbital period with horizontal boxes
        sns.boxplot(
            x=x_col_name,
            y=y_col_name,
            data=df,
            whis=[0, 100],
            width=0.6,
            fill=False,
            gap=0.1,
            color="grey",
            # palette="grey",
        )
        # Add in points to show each observation
        sns.stripplot(
            x=x_col_name,
            y=y_col_name,
            data=df,
            size=2,
            color="whitesmoke",
            linewidth=0.8,
            alpha=0.8,
        )
        # Tweak the visual presentation
        # ax.xaxis.grid(True)
        ax.set(ylabel="")
        ax.set(xlabel="")
        sns.despine(trim=True, left=True)

        ax.set_title(xlabel, fontsize=16)
        fig.subplots_adjust(
            # bottom=0.26,
            # top=0.92,
            left=0.25,
            # right=0.95, hspace=0.25, wspace=0.17
        )
        plt.show()


if __name__ == "__main__":

    p = editdistance()
    print(p.wk2())
    # print(p.asdas())

    # print(p.data())
# pd.options.display.multi_sparse = False

# # Define dataframes
# df1=pd.DataFrame(np.random.rand(4,2),index=["A","B","C","D"], columns=["zI","yJ"])
# df1.index.name='Criteria'

# df2=pd.DataFrame(np.random.rand(4,2),index=["A","B","C","D"], columns=["zI","yJ"])
# df2.index.name='Criteria'

# # Build comparison table
# comp = lsg.build_comparison_table([df1, df2], ['df1', 'df2'], fillna=0.0)

# ax, fig = lsg.plot_grouped_stackedbars(comp, ix_categories='Criteria', ix_entities_compared='Scenarios', norm=None)

# plt.show()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from umiche.fastq.Convert import convert as fas2bam
from umiche.trim.Template import template as umitrim
from umiche.util.Writer import writer as gwriter
from umiche.network.bfs.ConnectedComponent import connectedComponent as gbfscc
from umiche.deduplicate.pipeline import Config
from umiche.deduplicate.SingleCell import dedupSC
from umiche.plot.scenario.TraceSingle import valid as plotv
from umiche.path import to


class umi(Config.config):

    def __init__(self, metric, method, fastq_fp=None, is_trim=False, is_tobam=False, is_dedup=False):
        super(umi, self).__init__()
        self.metric = metric
        self.gbfscc = gbfscc()
        self.method = method
        self.gwriter = gwriter()
        self.plotv = plotv()
        df_dedup = pd.DataFrame()
        for i_pn in range(self.permutation_num):
            for id, i_metric in enumerate(self.metric_vals[self.metric]):
                if self.metric == 'pcr_nums':
                    print('=>at PCR {}'.format(i_metric))
                    fn_surf = str(i_metric)
                    self.mcl_inflat = i_metric
                    self.umi_len = self.umi_unit_len_fixed
                elif self.metric == 'pcr_errs':
                    self.mcl_inflat = i_metric
                    print('=>No.{} PCR error: {}'.format(id, i_metric))
                    fn_surf = str(id)
                    self.umi_len = self.umi_unit_len_fixed
                elif self.metric == 'seq_errs':
                    print('=>No.{} sequencing error: {}'.format(id, i_metric))
                    self.mcl_inflat = 1.4 if i_metric > 0.005 else 2.7
                    self.mcl_exp = 3
                    fn_surf = str(id)
                    self.umi_len = self.umi_unit_len_fixed
                elif self.metric == 'ampl_rates':
                    print('=>No.{} amplification rate: {}'.format(id, i_metric))
                    fn_surf = str(id)
                    # self.mcl_inflat = 1.3 if i_metric > 0.5 else 2
                    self.mcl_inflat = 2.3
                    self.mcl_exp = 2
                    self.umi_len = self.umi_unit_len_fixed
                elif self.metric == 'umi_lens':
                    print('=>No.{} UMI length: {}'.format(id, i_metric))
                    fn_surf = str(i_metric)
                    self.mcl_inflat = 1.1 if i_metric > 0.005 else 4
                    self.mcl_exp = 2
                    self.umi_len = i_metric
                else:
                    fn_surf = str(i_metric)
                    self.umi_len = self.umi_unit_len_fixed
                fn = self.fn_pref[self.metric] + fn_surf
                if is_trim:
                    self.trim(
                        fastq_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + fn,
                        fastq_trimmed_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/trimmed/' + fn,
                        umi_len=self.umi_len,
                    )
                if is_tobam:
                    fas2bam(
                        fastq_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/trimmed/' + fn + '.fastq.gz',
                        bam_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/bam/' + fn,
                    ).tobamsc()
                if is_dedup:
                    # if self.metric == 'seq_errs':
                        dedup_ob = dedupSC(
                            mode='internal',
                            method=self.method,

                            bam_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/bam/' + fn + '.bam',
                            gene_assigned_tag='XT',
                            gene_is_assigned_tag='XS',
                            mcl_fold_thres=1.6,
                            inflat_val=self.mcl_inflat,
                            exp_val=self.mcl_exp,
                            iter_num=100,
                            verbose=False,
                            ed_thres=1,
                            is_sv=False,
                            sv_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/summary/' + fn,
                        )
                        df_dedup.index = dedup_ob.dedup_index
                        df_dedup['metric' + str(id)] = dedup_ob.dedup_num
                        print(df_dedup)
        self.gwriter.generic(
            df=df_dedup,
            sv_fpn=fastq_fp + self.metric + '/' + str(self.method) + '.txt',
            header=True,
        )

    def trim(self, fastq_fpn, fastq_trimmed_fpn, umi_len):
        trim_params = {
            'read_struct': 'umi_1',
            'umi_1': {
                'len': umi_len,
            },
            'fastq': {
                'fpn': fastq_fpn + '.fastq.gz',
                'trimmed_fpn': fastq_trimmed_fpn + '.fastq.gz',
            },
        }
        umitrim_parser = umitrim(trim_params)
        df = umitrim_parser.todf()
        umitrim_parser.togz(df)
        return 0


if __name__ == "__main__":
    p = umi(
        # metric='pcr_nums',
        # metric='pcr_errs',
        metric='seq_errs',
        # metric='ampl_rates',
        # metric='umi_lens',

        # method='unique',
        # method='cluster',
        # method='adjacency',
        # method='directional',
        # method='mcl',
        # method='mcl_val',
        method='mcl_ed',

        # is_trim=True,
        # is_tobam=True,
        # is_dedup=False,

        is_trim=False,
        is_tobam=False,
        is_dedup=True,
        fastq_fp=to('data/simu/monomer/sc/'),
    )
    # print(p.evaluate())


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np


class Homotrimer:

    def __init__(self, ):
        pass

    @property
    def file_names(self, ):
        return {
            'pcr_nums': 'pcr_num_',
            'pcr_errs': 'pcr_err_',
            'seq_errs': 'seq_err_',
            'ampl_rates': 'ampl_rate_',
            'umi_lens': 'umi_len_',
            'seq_deps': 'seq_dep_',
            'umi_nums': 'umi_num_',
        }

    @property
    def fixed(self, ):
        return {
            'pcr_num': 8,
            'pcr_err': 1e-5,
            'seq_err': 1e-3,
            'ampl_rate': 0.85,
            'umi_len': 36,
            'seq_dep': 400,
            'umi_num': 50,
            'permutation_num': 10,
            'umi_unit_pattern': 3,
            'umi_unit_len': 12,
            'seq_sub_spl_rate': 1,
            'sim_thres': 3,
        }

    @property
    def varied(self, ):
        return {
            'pcr_nums': np.arange(1, 20 + 1, 1),
            'pcr_errs': self.errors(),
            'seq_errs': self.errors(),
            'ampl_rates': np.linspace(0.1, 1, 10),
            # 'umi_lens': np.arange(6, 36 + 3, 3),
            'umi_lens': np.arange(18, 60 + 3, 3),
            'seq_deps': np.arange(100, 1000 + 100, 100),
            'umi_nums': np.arange(50, 1000 + 200, 200),
        }

    def errors(self, ):
        arr = []
        e = 1e-5
        while e < 3e-1:
            arr.append(e)
            if 5 * e < 3e-1:
                arr.append(2.5 * e)
                arr.append(5 * e)
                arr.append(7.5 * e)
            e = 10 * e
        arr.append(0.125)
        arr.append(0.15)
        arr.append(0.2)
        arr.append(0.225)
        arr.append(0.25)
        arr.append(0.3)
        return arr

    def errs(self, ):
        return [
            1e-05,
            2.5e-05,
            5e-05,
            7.5e-05,
            0.0001,
            0.00025,
            0.0005,
            0.00075,
            0.001,
            0.0025,
            0.005,
            0.0075,
            0.01,
        ]


if __name__ == "__main__":
    p = Homotrimer()

    print(p.file_names)

    print(p.varied)

    print(p.fixed)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"



class Hamming:

    def general(self, s1, s2):
        """

        Parameters
        ----------
        s1
            sequence 1
        s2
            sequence 2

        Returns
        -------
            int

        """
        return sum(i != j for i, j in zip(s1, s2))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from datetime import datetime


class Console:
    
    def __init__(self, placeholder='logger: ', verbose=False):
        self._verbose = verbose
        self.placeholder = placeholder

    @property
    def verbose(self, ):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    def print(self, content):
        if self._verbose:
            now = datetime.now()
            dt_format = now.strftime("%d/%m/%Y %H:%M:%S ")
            print(dt_format + self.placeholder + str(content))

    def check(self, content):
        if self._verbose:
            print(content)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import pandas as pd
from path import to
from ext.sequelpy.plot.scatter.DimensionReduction import dimensionReduction as drplot


def read(
        gmat_fpn,
):
    df_gmat = pd.read_hdf(gmat_fpn, 'df')
    y = df_gmat['Y'].values
    if 'Y' in df_gmat.columns:
        df_gmat = df_gmat.drop(columns=['Y'])
    return df_gmat, y


# title = '10 epoch'
# title = '20 epoch'
# title = '50 epoch'
title = '100 epoch'

df, y = read(
    gmat_fpn=to('umiche/') + title + '.h5',
)
x = df.values
# print(x)
# print(y)

drplot().single(X=x, y=y, tech='PCA', marker_size=3, cmap='Set3', title=title)
drplot().single(X=x, y=y, tech='UMAP', marker_size=3, cmap='Set3', title=title)
drplot().single(X=x, y=y, tech='TSNE', marker_size=3, cmap='Set3', title=title) # Set3 tab20b



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__ = "jianfeng.sunmt@gmail.com"


import time

from umiche.bam.Writer import Writer as aliwriter

from umiche.deduplicate.Gadgetry import Gadgetry as umigadgetry

# dedup methods
from umiche.deduplicate.method.Adjacency import Adjacency as umiadj
from umiche.deduplicate.method.Directional import Directional as umidirec
from umiche.deduplicate.method.MarkovClustering import MarkovClustering as umimcl
from umiche.deduplicate.method.Clustering import Clustering as umiclustering
from umiche.deduplicate.method.trimer.SetCover import SetCover as umisc
from umiche.deduplicate.method.trimer.MajorityVote import MajorityVote as umimv

from umiche.util.Writer import Writer as fwriter
from umiche.util.Console import Console


class Tabulate:

    def __init__(
            self,
            df,
            df_bam,
            bam_fpn,
            work_dir,
            heterogeneity,
            verbose=False,
    ):
        self.df = df
        self.df_bam = df_bam
        self.bam_fpn = bam_fpn
        self.work_dir = work_dir
        self.heterogeneity = heterogeneity

        self.aliwriter = aliwriter(df=self.df_bam)

        self.umigadgetry = umigadgetry()

        self.fwriter = fwriter()

        self.console = Console()
        self.console.verbose = verbose

    def set_cover(
            self,
            **kwargs,
    ):
        self.df_umi_uniq = self.df_bam.drop_duplicates(subset=['umi'], keep='first')
        # print(self.df_umi_uniq)
        series_uniq_umi = self.df_umi_uniq.umi
        # print(series_uniq_umi)
        self.umi_to_int_dict = {k: id for id, k in enumerate(series_uniq_umi)}

        dedup_cnt, multimer_umi_solved_by_sc, multimer_umi_not_solved, shortlisted_multimer_umi_list, monomer_umi_lens, multimer_umi_lens = umisc().greedy(
            multimer_list=series_uniq_umi.values,
            recur_len=kwargs['umi_unit_pattern'],
            split_method=kwargs['split_method'],
        )
        # print(monomer_umi_lens)
        # print(multimer_umi_lens)
        # print(dedup_cnt)
        self.df.loc[0, 'dedup_cnt'] = dedup_cnt
        self.df.loc[0, 'num_solved'] = len(multimer_umi_solved_by_sc)
        self.df.loc[0, 'num_not_solved'] = len(multimer_umi_not_solved)
        self.df.loc[0, 'monomer_umi_len'] = ';'.join([str(i) for i in monomer_umi_lens])
        self.df.loc[0, 'multimer_umi_len'] = ';'.join([str(i) for i in multimer_umi_lens])

        sc_bam_ids = []
        for i in shortlisted_multimer_umi_list:
            sc_bam_ids.append(series_uniq_umi.loc[series_uniq_umi.isin([i])].index[0])

        self.console.print('======>start writing deduplicated reads to BAM...')
        dedup_reads_write_stime = time.time()
        # print(self.work_dir)

        import os
        from umiche.util.Folder import Folder as crtfolder
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(kwargs['sv_interm_bam_fpn']))

        self.aliwriter.tobam(
            tobam_fpn=kwargs['sv_interm_bam_fpn'],
            tmpl_bam_fpn=self.bam_fpn,
            whitelist=sc_bam_ids,
        )
        self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def majority_vote(
            self,
            **kwargs,
    ):
        self.df_umi_uniq = self.df_bam.drop_duplicates(subset=['umi'], keep='first')
        # print(self.df_umi_uniq)
        series_uniq_umi = self.df_umi_uniq.umi
        # print(series_uniq_umi)
        self.umi_to_int_dict = {k: id for id, k in enumerate(series_uniq_umi)}

        dedup_cnt, uniq_multimer_cnt, shortlisted_multimer_umi_list = umimv().track(
            multimer_list=series_uniq_umi.values,
            recur_len=kwargs['umi_unit_pattern'],
        )
        # print(dedup_cnt)
        self.df.loc[0, 'dedup_cnt'] = dedup_cnt
        sc_bam_ids = []
        for i in shortlisted_multimer_umi_list:
            sc_bam_ids.append(series_uniq_umi.loc[series_uniq_umi.isin([i])].index[0])

        self.console.print('======>start writing deduplicated reads to BAM...')
        dedup_reads_write_stime = time.time()
        # print(self.work_dir)

        import os
        from umiche.util.Folder import Folder as crtfolder
        crtfolder().osmkdir(DIRECTORY=os.path.dirname(kwargs['sv_interm_bam_fpn']))

        self.aliwriter.tobam(
            tobam_fpn=kwargs['sv_interm_bam_fpn'],
            tmpl_bam_fpn=self.bam_fpn,
            whitelist=sc_bam_ids,
        )
        self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def unique(self, ):
        dedup_umi_stime = time.time()
        self.df['uniq_sgl_mark'] = self.df['uniq_repr_nodes'].apply(lambda x: 'yes' if len(x) == 1 else 'no')
        self.df = self.df.loc[self.df['uniq_sgl_mark'] == 'no']
        self.console.print('======># of positions with non-single umis: {}'.format(self.df.shape[0]))
        self.console.print('======>finish finding deduplicated UMIs in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of UMIs deduplicated {}'.format(self.df['num_uniq_umis'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        ### @@ self.df['ave_ed']
        # 1    5.0
        # Name: ave_eds, dtype: float64
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='uniq_repr_nodes'), axis=1)
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='uniq_repr_nodes'), axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col='uniq_repr_nodes'),
            axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'unique_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'unique_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['uniq_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='uniq_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'unique_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['uniq_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def cluster(self, ):
        dedup_umi_stime = time.time()
        self.df['cc_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='cc'), axis=1)
        ### @@ self.df['cc_repr_nodes']
        # 1    [2]
        # Name: cc_repr_nodes, dtype: object
        self.df['dedup_cnt'] = self.df['cc_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        ### @@ self.df['dedup_cnt']
        # 1    1
        # Name: dedup_cnt, dtype: int64
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='cc_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='cc_repr_nodes'),
            axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col='cc_repr_nodes'),
            axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'cluster_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'cluster_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['cc_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='cc_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'cluster_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['cc_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def adjacency(self, ):
        dedup_umi_stime = time.time()
        umiadj_ob = umiadj()
        self.df['adj'] = self.df.apply(
            lambda x: umiadj_ob.decompose(
                cc_sub_dict=umiadj_ob.umi_tools(
                    connected_components=x['cc'],
                    df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                    graph_adj=x['vignette']['graph_adj'],
                )['clusters'],
            ),
            axis=1,
        )
        self.df['adj_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='adj'), axis=1)
        self.df['dedup_cnt'] = self.df['adj_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='adj_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(
                x,
                by_col='adj_repr_nodes',
            ),
            axis=1,
        )
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(
                x,
                by_col='adj_repr_nodes',
            ),
            axis=1,
        )
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'adjacency_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'adjacency_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['adj_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='adj_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'adjacency_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['adj_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def directional(self, ):
        dedup_umi_stime = time.time()
        umidirec_ob = umidirec(self.heterogeneity)
        # self.df[['count', 'clusters', 'apv', 'disapv']] = self.df.apply(
        #     lambda x: umidirec_ob.umi_tools(
        #         connected_components=x['cc'],
        #         df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
        #         graph_adj=x['vignette']['graph_adj'],
        #     ),
        #     axis=1,
        #     result_type='expand',
        # )
        if self.heterogeneity:
            self.df['count'], self.df['clusters'], self.df['apv'], self.df['disapv'] = zip(
                *self.df.apply(
                    lambda x: umidirec_ob.umi_tools(
                        connected_components=x['cc'],
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        graph_adj=x['vignette']['graph_adj'],
                    ),
                    axis=1,
                )
            )
            self.df['direc'] = self.df.apply(
                lambda x: umidirec_ob.decompose(
                    cc_sub_dict=x['clusters'],
                ),
                axis=1,
            )
        else:
            self.df['direc'] = self.df.apply(
                lambda x: umidirec_ob.decompose(
                    cc_sub_dict=umidirec_ob.umi_tools(
                        connected_components=x['cc'],
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        graph_adj=x['vignette']['graph_adj'],
                    )['clusters'],
                ),
                axis=1,
            )
        # print(self.df['direc'])
        self.df['direc_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='direc'), axis=1)
        # print(self.df['direc_repr_nodes'])
        self.df['dedup_cnt'] = self.df['direc_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='direc_repr_nodes'), axis=1)
        # print(self.df['ave_eds'])
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='direc_repr_nodes'),
            axis=1,
        )
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col='direc_repr_nodes'),
            axis=1,
        )
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        # print(self.ave_ed_bins)
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'directional_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'directional_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['direc_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='direc_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'directional_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['direc_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def mcl(
            self,
            **kwargs
    ):
        # print(self.df.columns)
        # print(self.df)
        dedup_umi_stime = time.time()
        umimcl_ob = umimcl(
            inflat_val=kwargs['inflat_val'],
            exp_val=kwargs['exp_val'],
            iter_num=kwargs['iter_num'],
            heterogeneity=self.heterogeneity,
        )
        # print(self.df)
        ### @@ please note that df and df_mcl_res cannot be merged becuase the dimension is not the same.
        if self.heterogeneity:
            df_mcl_res = self.df.apply(
                lambda x: umimcl_ob.dfclusters(
                    connected_components=x['cc'],
                    graph_adj=x['vignette']['graph_adj'],
                ),
                axis=1,
            ).values[0]
            mcl_dict = {'mcl': umimcl_ob.decompose(
                list_nd=df_mcl_res['clusters'].values
            )}
            self.df['mcl'] = self.df.apply(lambda x: mcl_dict['mcl'], axis=1)
            apv_dict = {'apv': [df_mcl_res['apv']]}
            self.df['apv'] = self.df.apply(lambda x: apv_dict['apv'], axis=1)
            # print(self.df['mcl'])
        else:
            self.df['mcl'] = self.df.apply(
                lambda x: umimcl_ob.decompose(
                    list_nd=umimcl_ob.dfclusters(
                        connected_components=x['cc'],
                        graph_adj=x['vignette']['graph_adj'],
                    )['clusters'].values,
                ),
                axis=1,
            )
            # print(self.df['mcl'])
        ### @@ self.df['mcl']
        # 1    {0: [0, 76, 162, 188, 237, 256], 1: [65, 55, 1...
        # Name: mcl, dtype: object
        self.df['mcl_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='mcl'), axis=1)
        self.df['dedup_cnt'] = self.df['mcl_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='mcl_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='mcl_repr_nodes'),axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(lambda x: self.umigadgetry.num_removed_reads(x, by_col='mcl_repr_nodes'),axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'mcl_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'mcl_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['mcl_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='mcl_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'mcl_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['mcl_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def mcl_cc_all_node_umis(
            self,
            **kwargs,
    ):
        dedup_umi_stime = time.time()
        umimcl_ob = umimcl(
            inflat_val=kwargs['inflat_val'],
            exp_val=kwargs['exp_val'],
            iter_num=kwargs['iter_num'],
            heterogeneity=self.heterogeneity,
        )
        # print(self.df)
        ### @@ please note that df and df_mcl_res cannot be merged becuase the dimension is not the same.
        df_mcl_res = self.df.apply(
            lambda x: umimcl_ob.dfclusters_cc_all_node_umis(
                int_to_umi_dict=x['vignette']['int_to_umi_dict'],
                graph_adj=x['vignette']['graph_adj'],
            ),
            axis=1,
        ).values[0]
        mcl_dict = {'mcl': umimcl_ob.decompose(
            list_nd=df_mcl_res['clusters'].values
        )}
        apv_dict = {'apv': [df_mcl_res['apv']]}
        self.df['mcl'] = self.df.apply(lambda x: mcl_dict['mcl'], axis=1)
        self.df['apv'] = self.df.apply(lambda x: apv_dict['apv'], axis=1)
        # print(self.df['apv'])
        # self.df['mcl'] = self.df.apply(
        #     lambda x: umimcl_ob.decompose(
        #         list_nd=umimcl_ob.dfclusters(
        #             connected_components=x['cc'],
        #             graph_adj=x['vignette']['graph_adj'],
        #         )['clusters'].values,
        #     ),
        #     axis=1,
        # )
        ### @@ self.df['mcl']
        # 1    {0: [0, 76, 162, 188, 237, 256], 1: [65, 55, 1...
        # Name: mcl, dtype: object
        self.df['mcl_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='mcl'), axis=1)
        self.df['dedup_cnt'] = self.df['mcl_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='mcl_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='mcl_repr_nodes'),axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(lambda x: self.umigadgetry.num_removed_reads(x, by_col='mcl_repr_nodes'),axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'mcl_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'mcl_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['mcl_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='mcl_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'mcl_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['mcl_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def mcl_val(
            self,
            **kwargs,
    ):
        dedup_umi_stime = time.time()
        umimcl_ob = umimcl(
            inflat_val=kwargs['inflat_val'],
            exp_val=kwargs['exp_val'],
            iter_num=kwargs['iter_num'],
            heterogeneity=self.heterogeneity,
        )
        if self.heterogeneity:
            self.df['count'], self.df['clusters'], self.df['apv'], self.df['disapv'] = zip(
                *self.df.apply(
                    lambda x: umimcl_ob.maxval_val(
                        df_mcl_ccs=umimcl_ob.dfclusters(
                            connected_components=x['cc'],
                            graph_adj=x['vignette']['graph_adj'],
                        ),
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        thres_fold=kwargs['mcl_fold_thres'],
                    ),
                    axis=1,
                )
            )
            self.df['mcl_val'] = self.df.apply(
                lambda x: umimcl_ob.decompose(
                    list_nd=x['clusters'].values,
                ),
                axis=1,
            )
            # print(self.df['mcl_val'])
        else:
            self.df['mcl_val'] = self.df.apply(
                lambda x: umimcl_ob.decompose(
                    list_nd=umimcl_ob.maxval_val(
                        df_mcl_ccs=umimcl_ob.dfclusters(
                            connected_components=x['cc'],
                            graph_adj=x['vignette']['graph_adj'],
                        ),
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        thres_fold=kwargs['mcl_fold_thres'],
                    )['clusters'].values,
                ),
                axis=1,
            )
            # print(self.df['mcl_val'])
        self.df['mcl_val_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='mcl_val'), axis=1)
        self.df['dedup_cnt'] = self.df['mcl_val_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='mcl_val_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='mcl_val_repr_nodes'), axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col='mcl_val_repr_nodes'), axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'mcl_val_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'mcl_val_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['mcl_val_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='mcl_val_repr_nodes'),
                                                       axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'mcl_val_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['mcl_val_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def mcl_ed(
            self,
            **kwargs,
    ):
        dedup_umi_stime = time.time()
        umimcl_ob = umimcl(
            inflat_val=kwargs['inflat_val'],
            exp_val=kwargs['exp_val'],
            iter_num=kwargs['iter_num'],
            heterogeneity=self.heterogeneity,
        )
        if self.heterogeneity:
            # self.df[['count', 'clusters', 'apv', 'disapv']]
            self.df['count'], self.df['clusters'], self.df['apv'], self.df['disapv'] = zip(
                *self.df.apply(
                    lambda x: umimcl_ob.maxval_ed(
                        df_mcl_ccs=umimcl_ob.dfclusters(
                            connected_components=x['cc'],
                            graph_adj=x['vignette']['graph_adj'],
                        ),
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        int_to_umi_dict=x['vignette']['int_to_umi_dict'],
                        thres_fold=kwargs['mcl_fold_thres'],
                    ),
                    axis=1,
                )
            )
            self.df['mcl_ed'] = self.df.apply(
                lambda x: umimcl_ob.decompose(
                    list_nd=x['clusters'].values,
                    # list_nd=x['clusters'].values[0].tolist(),
                ),
                axis=1,
            )
            # print(self.df['mcl_ed'])
        else:
            self.df['mcl_ed'] = self.df.apply(
                lambda x: umimcl_ob.decompose(
                    list_nd=umimcl_ob.maxval_ed(
                        df_mcl_ccs=umimcl_ob.dfclusters(
                            connected_components=x['cc'],
                            graph_adj=x['vignette']['graph_adj'],
                        ),
                        df_umi_uniq_val_cnt=x['vignette']['df_umi_uniq_val_cnt'],
                        int_to_umi_dict=x['vignette']['int_to_umi_dict'],
                        thres_fold=kwargs['mcl_fold_thres'],
                    )['clusters'].values,
                ),
                axis=1,
            )
        self.df['mcl_ed_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col='mcl_ed'), axis=1)
        self.df['dedup_cnt'] = self.df['mcl_ed_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        # print(self.df['dedup_cnt'])
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col='mcl_ed_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col='mcl_ed_repr_nodes'), axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col='mcl_ed_repr_nodes'),
            axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + 'mcl_ed_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + 'mcl_ed_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df['mcl_ed_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col='mcl_ed_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + 'mcl_ed_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df['mcl_ed_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df

    def clustering_umi_seq_onehot(
            self,
            clustering_method,
            **kwargs
    ):
        dedup_umi_stime = time.time()
        self.umiclustering = umiclustering(
            clustering_method=clustering_method,
            **kwargs,
        )
        df_clustering_res = self.df.apply(
            lambda x: self.umiclustering.dfclusters(
                connected_components=x['cc'],
                graph_adj=x['vignette']['graph_adj'],
                int_to_umi_dict=x['vignette']['int_to_umi_dict'],
            ),
            axis=1,
        ).values[0]
        mcl_dict = {clustering_method: self.umiclustering.decompose(
            list_nd=df_clustering_res['clusters'].values
        )}
        apv_dict = {'apv': [df_clustering_res['apv']]}
        self.df[clustering_method] = self.df.apply(lambda x: mcl_dict[clustering_method], axis=1)
        self.df['apv'] = self.df.apply(lambda x: apv_dict['apv'], axis=1)
        # print(self.df['apv'])
        self.df[clustering_method + '_repr_nodes'] = self.df.apply(lambda x: self.umigadgetry.umimax(x, by_col=clustering_method), axis=1)
        self.df['dedup_cnt'] = self.df[clustering_method + '_repr_nodes'].apply(lambda x: self.umigadgetry.length(x))
        self.console.print('======>finish finding deduplicated umis in {:.2f}s'.format(time.time() - dedup_umi_stime))
        # self.console.print('======># of umis deduplicated to be {}'.format(self.df['dedup_cnt'].loc['yes']))
        self.console.print('======>calculate average edit distances between umis...')
        self.df['ave_eds'] = self.df.apply(lambda x: self.umigadgetry.ed_ave(x, by_col=clustering_method + '_repr_nodes'), axis=1)
        self.df['num_diff_dedup_uniq_umis'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_uniq_umis(x, by_col=clustering_method + '_repr_nodes'),
            axis=1)
        self.df['num_diff_dedup_reads'] = self.df.apply(
            lambda x: self.umigadgetry.num_removed_reads(x, by_col=clustering_method + '_repr_nodes'),
            axis=1)
        self.console.print('======># of deduplicated unique umis {}'.format(self.df['num_diff_dedup_uniq_umis'].sum()))
        self.console.print('======># of deduplicated reads {}'.format(self.df['num_diff_dedup_reads'].sum()))
        self.ave_ed_bins = self.df['ave_eds'].value_counts().sort_index().to_frame().reset_index()
        self.console.check("======>bins for average edit distance\n{}".format(self.ave_ed_bins))
        if not self.heterogeneity:
            self.fwriter.generic(
                df=self.ave_ed_bins,
                sv_fpn=self.work_dir + clustering_method + '_ave_ed_bin.txt',
                index=True,
                header=True,
            )
            self.fwriter.generic(
                df=self.df[[
                    'dedup_cnt',
                    'ave_ed',
                    'num_uniq_umis',
                    'num_diff_dedup_uniq_umis',
                    'num_diff_dedup_reads',
                ]],
                sv_fpn=self.work_dir + clustering_method + '_dedup_sum.txt',
                index=True,
                header=True,
            )
            self.console.print('======>start writing deduplicated reads to BAM...')
            dedup_reads_write_stime = time.time()
            self.df[clustering_method + '_bam_ids'] = self.df.apply(lambda x: self.umigadgetry.bamids(x, by_col=clustering_method + '_repr_nodes'), axis=1)
            self.aliwriter.tobam(
                tobam_fpn=self.work_dir + clustering_method + '_dedup.bam',
                tmpl_bam_fpn=self.bam_fpn,
                whitelist=self.umigadgetry.decompose(list_nd=self.df[clustering_method + '_bam_ids'].values),
            )
            self.console.print('======>finish writing in {:.2f}s'.format(time.time() - dedup_reads_write_stime))
        return self.df